/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.calculation.calculation.typeNames;

import java.util.HashSet;
import java.util.Set;

import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaModelException;
import org.intoJ.inferType.model.model.DeclarationElement;

/**
 * This class is the abstract superclass of type name generators.
 *
 * @see org.intoJ.inferType.calculation.calculation.typeNames.DefaultTypeNameGenerator
 *  
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Andreas Meiner
 */
public abstract class TypeNameGenerator {

	private IJavaProject javaProject;
	protected static Set usedNames = new HashSet();

	/**
	 * The constructor.
	 * 
	 * @param javaProject the current IJavaProject
	 */
	public TypeNameGenerator(IJavaProject javaProject, boolean reset) {

		this.javaProject = javaProject;
		if (reset)
			reset();
	}

	/**
	 * This method generates the type name. The new type depends on the given
	 * declaration element. 
	 * 
	 * @param de DeclarationElement
	 * @return full qualified type name
	 */
	public abstract String getName(DeclarationElement de);

	/**
	 * This method determines the default package for all following created type
	 * names.
	 * 
	 * @param de the declaration element infer type was executed on
	 * @return the default package
	 */
	protected abstract String determinePackage(DeclarationElement de);

	/**
	 * This method clears the used type names set.
	 */
	public static void reset() {

		usedNames = new HashSet();
	}

	/**
	 * This method checks whether the new type name is a valid one.
	 * It is valid if
	 * - the name does not belong to a type already
	 * - the name has not been created yet by this class
	 * - the name does not correspond to a file of the same name in the platform
	 *      (for example, on Windows, lower and upper case doesn't matter)
	 * 
	 * @param fqTypeName the potential new full qualified type name
	 * @return true if the new type name is valid
	 */
	protected boolean isValid(String thePackage, String theSimpleName) {

		try {

			if (usedNames.contains(theSimpleName))
				return false;
			if (javaProject.findType(thePackage + "." + theSimpleName) != null)
				return false;

			return true;

		} catch (JavaModelException e) {
			return false;
		}
	}
}
