/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.model.model;

import java.util.Iterator;

import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.intoJ.inferType.model.collection.DESet_Key;
import org.intoJ.inferType.model.collection.DESet_Signature;
import org.intoJ.inferType.model.collection.TypeSet;
import org.intoJ.inferType.util.Util_TypeNames;

/**
 * 
 * Wrapper for ITypeBinding.
 * 
 * FAQ: 
 * 
 * Q: Why do we need a wrapper for ITypeBinding? 
 * A: ITypeBinding only contains declared methods, not inherited ones. We need those.
 * 
 * Q: Why not take an IType ? 
 * A: ITypes are not always available while parsing  source, they also do not contain the fully qualified 
 *    names of method return types, parameters, etc. (just "QString" f.e.). Also, this class 
 *    features correct fully qualified name for local types (IType does not).
 * 
 * NOTE! As of revision 1.2, TypeElement is change-aware, meaning it contains fields and methods for handling
 *       newly created types or changed types (a changed type containing lesser, or "superfluous", methods) as
 *       well as a collection of DEs which will be re-typed to this type.
 *  	 The binding is no longer guaranteed to not be null. 
 * 
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 */
public class TypeElement extends InformingObject {

	/**
	 * The binding corresponding to this type element. The binding may be null,
	 * indicating that the element does not yet exist.
	 */
	private ITypeBinding binding = null;
	/**
	 * non-qualified type name
	 */
	private String fqTypeName = "";
	/**
	 * If true, this element is an interface.
	 */
	private boolean isInterface = false;
	/**
	 * If true, this element may be a class if it so chooses.
	 */
	private boolean mayBeClass = false;
	/**
	 * If true, this element may be an interface if it so chooses.
	 */
	private boolean mayBeInterface = true;
	/**
	 * If true, this element does not yet exist.
	 */
	private boolean isNewTypeElement = false;
	/**
	 * Contains all inherited members (fields, methods)
	 */
	private DESet_Signature inheritedMembers = new DESet_Signature();
	/**
	 * Contains all declared members (non-inherited members, fields and methods) 
	 */
	private DESet_Signature declaredMembers = new DESet_Signature();
	/**
	 * Superfluous members in the member list.
	 */
	private DESet_Signature superfluousMembers = new DESet_Signature();
	/**
	 * Contains inherited (non-direct) supertypes of this type.
	 */
	private TypeSet inheritedSupertypes = new TypeSet();
	/**
	 * Contains the direct supertypes of this type.
	 */
	private TypeSet declaredSupertypes = new TypeSet();
	/**
	 * Contains the DeclarationElements which will be typed with this
	 * type after completion (for information only)
	 */
	private DESet_Signature desAffectedByThisType = new DESet_Signature();
	/**
	 * This is actually very bad design ;)
	 */
	private boolean openInEditor = false;
	private boolean isUsedForRootElement = false;

	/**
	 * Constructor for existing types with bindings. All fields are initialized.
	 * 
	 * @param binding Binding for this element. May be null.
	 */
	public TypeElement(ITypeBinding binding) {

		this.binding = binding;
		setIsNewType(false);
		this.isInterface = binding.isInterface();
		this.fqTypeName = Util_TypeNames.getQualifiedNameForAllTypesOfTypes(binding);
	}

	/**
	 * Constructor for types without bindings. All fields are initialized.
	 * @param fqName
	 * @param isInterface
	 * @param isNewType
	 */
	public TypeElement(String fqName, boolean isInterface, boolean isNewType) {

		this.binding = null;
		setIsNewType(isNewType);
		this.isInterface = isInterface;
		this.fqTypeName = fqName;
	}

	//************************************ TYPE ITSELF *********************************

	/**
	 * Returns the binding of this type.
	 * 
	 * @return ITypeBinding
	 */
	public ITypeBinding getBinding() {

		return binding;
	}

	/**
	 * Returns the fully qualified name of this type.
	 * 
	 * @return String
	 */
	public String getFullyQualifiedName() {

		return fqTypeName;
	}

	/**
	 * True if this type is an interface.
	 * 
	 * @return boolean
	 */
	public boolean isInterface() {

		return isInterface;
	}

	/**
	 * True if this is a new type.
	 * 
	 * @return boolean
	 */
	public boolean isNewType() {

		return isNewTypeElement;
	}

	/**
	 * True if this interface may also be a class.
	 * 
	 * @return boolean
	 */
	public boolean isMayBeClass() {

		return mayBeClass;
	}

	/**
	 * Sets the maybeclass field.
	 * 
	 * @param mayBeClass boolean
	 */
	public void setMayBeClass(boolean mayBeClass) {

		this.mayBeClass = mayBeClass;
		if (!isInterface && !mayBeClass)
			isInterface = true;
	}

	/**
	 * Sets the maybeinterface field.
	 * 
	 * @param mayBeInterface boolean
	 */
	public void setMayBeInterface(boolean mayBeInterface) {

		this.mayBeInterface = mayBeInterface;
		if (isInterface && !mayBeInterface)
			isInterface = false;
	}

	/**
	 * True if this class may be an interface.
	 * 
	 * @return boolean
	 */
	public boolean isMayBeInterface() {

		return mayBeInterface;
	}

	/**
	 * Sets the isnewtype field.
	 * 
	 * @param b boolean
	 */
	public void setIsNewType(boolean b) {

		isNewTypeElement = b;
		//As per request 2005-04-25
		//openInEditor = isNewTypeElement;
	}

	/**
	 * True if this type is an abstract class or an interface
	 * @return boolean
	 */
	public boolean isAbstract() {

		if (isNewTypeElement)
			return true;
		if (isInterface)
			return true;
		if (binding == null)
			return false;

		return Modifier.isAbstract(binding.getModifiers());
	}

	/**
	 * Sets the fully qualified name (used for new types)
	 * 
	 * @param fqTypeName String
	 */
	public void setFullyQualifiedName(String fqTypeName) {

		if (!this.fqTypeName.equals(fqTypeName)) {
			this.fqTypeName = fqTypeName;
			firePropertyChange(TYPE_NAME, null, fqTypeName);
		}
	}

	/**
	 * Sets the isInterface field (used for new types).
	 * @param isInterface boolean
	 */
	public void setIsInterface(boolean isInterface) {

		if (!this.isInterface == isInterface) {
			this.isInterface = isInterface;
			firePropertyChange(TYPE_TYPE, null, new Boolean(isInterface));
		}
	}

	/**
	 * True if this type is to be opened in the editor.
	 * 
	 * @return boolean
	 */
	public boolean isOpenInEditor() {

		return openInEditor;
	}

	/**
	 * Sets the openInEditor field.
	 * 
	 * @param openInEditor boolean
	 */
	public void setOpenInEditor(boolean openInEditor) {

		this.openInEditor = openInEditor;
	}

	//************************************** MEMBERS ***********************************

	/**
	 * @return Returns all members, inherited and non-inherited ones
	 */
	public DESet_Signature getAllMembers() {

		DESet_Signature tmp = new DESet_Signature();
		tmp.addDeclarations(inheritedMembers);
		tmp.addDeclarations(declaredMembers);
		return tmp;
	}

	/**
	 * @return Returns the inherited members.
	 */
	public DESet_Signature getInheritedMembers() {

		return inheritedMembers;
	}

	/**
	 * @return Returns the own (non-inherited) members.
	 */
	public DESet_Signature getDeclaredMembers() {

		return declaredMembers;
	}

	/**
	 * @return Returns the superflous members (no longer needed, as they are defined
	 * in a superclass or interface).
	 */
	public DESet_Signature getSuperfluousMembers() {

		return superfluousMembers;
	}

	/**
	 * Adds inherited members (used for new types).
	 * @param declarationElements DESet_Signature
	 */
	public void addInheritedMembers(DESet_Signature declarationElements) {

		inheritedMembers.addDeclarations(declarationElements);
	}

	/**
	 * Adds declared members (used for new types).
	 * @param declarationElements DESet_Signature
	 */
	public void addDeclaredMembers(DESet_Signature declarationElements) {

		declaredMembers.addDeclarations(declarationElements);
	}

	/**
	 * Adds superfluous members (used for new types).
	 * @param declarationElements DESet_Signature
	 */
	public void addSuperfluousMembers(DESet_Signature declarationElements) {

		//Bugfix: Some members may not be present in this type.

		for (Iterator i = declarationElements.deIterator(); i.hasNext();) {
			DeclarationElement current = (DeclarationElement) i.next();
			if (this.getDeclaredMembers().containsDeclaration(current))
				superfluousMembers.addDeclaration(current);

		}
	}

	/**
	 * Returns the count of all members.
	 * 
	 * @return int
	 */
	public int allMemberCount() {

		return getAllMembers().size();
	}

	/**
	 * Checks whether all members in the list are contained in this element.
	 * @param members DESet_Key
	 * @return boolean
	 */
	public boolean containsAllMembers(DESet_Key members) {

		return getAllMembers().containsDeclarations(members);
	}

	//*********************************** SUPERTYPES ***********************************

	/**
	 * Returns all (directly or indirectly implemented) supertypes.
	 * @return TypeSet
	 */
	public TypeSet getAllSuperTypes() {

		TypeSet tmp = new TypeSet();
		tmp.addTypes(declaredSupertypes);
		tmp.addTypes(inheritedSupertypes);
		return tmp;
	}

	/**
	 * Returns all declared (direct) supertypes
	 * @return TypeSet
	 */
	public TypeSet getDeclaredSupertypes() {

		return declaredSupertypes;
	}

	/**
	 * Returns all inherited (non-direct) supertypes.
	 * @return TypeSet
	 */
	public TypeSet getInheritedSupertypes() {

		return inheritedSupertypes;
	}

	/**
	 * Adds a new declared supertype.
	 * @param type TypeElement
	 */
	public void addDeclaredSupertype(TypeElement type) {

		declaredSupertypes.addType(type);
	}

	/**
	 * Adds a new inherited supertype.
	 * @param types TypeSet
	 */
	public void addInheritedSupertype(TypeSet types) {

		inheritedSupertypes.addTypes(types);
	}

	/**
	 * Removes a declared supertype.
	 * @param type TypeElement
	 * @return boolean
	 */
	public boolean removeDeclaredSupertype(TypeElement type) {

		return declaredSupertypes.removeType(type);
	}

	/**
	 * Returns true if el is a direct supertype of this element.
	 * @param el TypeElement
	 * @return boolean
	 */
	public boolean hasDirectSupertype(TypeElement el) {

		return declaredSupertypes.contains(el);
	}

	/**
	 * Returns true if this element has a superclass.
	 * 
	 * @return boolean
	 */
	public boolean hasSuperclass() {

		for (Iterator i = declaredSupertypes.getAsList().iterator(); i.hasNext();) {
			TypeElement current = ((TypeElement) i.next());
			if (!current.isInterface() && !current.getFullyQualifiedName().equals("java.lang.Object"))
				return true;
		}
		return false;
	}

	/**
	 * Returns true if el is a direct or indirect supertype of this type.
	 * 
	 * @param el TypeElement
	 * @return boolean
	 */
	public boolean hasSupertype(TypeElement el) {

		return hasDirectSupertype(el) || inheritedSupertypes.contains(el);
	}

	//*********************************** OTHER STUFF ******************************************

	/**
	 * Normally, elements are compared using their key. However, some elements
	 * do not possess such a key (for example, class definitions inside another
	 * class).
	 * 
	 * @param o Object
	 * @return boolean
	 */
	public boolean equals(Object o) {

		if (o instanceof TypeElement) {
			TypeElement other = (TypeElement) o;
			if ((this.binding == null) || (other.binding == null))
				return getFullyQualifiedName().equals(other.getFullyQualifiedName());
			if ((other.binding.getKey() == null) || (this.binding.getKey() == null))
				return Util_TypeNames.getQualifiedNameForAllTypesOfTypes(other.binding).equals(Util_TypeNames.getQualifiedNameForAllTypesOfTypes(this.binding));
			return other.binding.getKey().equals(this.binding.getKey());
		}

		return false;
	}

	/**
	 * Method hashCode.
	 * @return int
	 */
	public int hashCode() {

		return (binding == null) ? getFullyQualifiedName().hashCode() : (binding.getKey() == null) ? binding.getQualifiedName().hashCode() : binding.getKey()
				.hashCode();
	}

	/**
	 * Method toString.
	 * @return String
	 */
	public String toString() {

		return getFullyQualifiedName();
	}

	/**
	 * Sets the type binding (used for new types).
	 * @param typeBinding
	 */
	public void setBinding(ITypeBinding typeBinding) {

		this.binding = typeBinding;

	}

	/**
	 * Returns the Declaration Elements declared with this type.
	 * 
	 * @return Returns the desAffectedByThisType.
	 */
	public DESet_Signature getDesAffectedByThisType() {

		return desAffectedByThisType;
	}

	/**
	 * Add an affected Declaration Element for this type.
	 * @param de DeclarationElement
	 */
	public void addAffectedDE(DeclarationElement de) {

		desAffectedByThisType.addDeclaration(de);

	}

	/**
	 * @return Returns the isUsedForRootElement.
	 */
	public boolean isUsedForRootElement() {

		return isUsedForRootElement;
	}

	/**
	 * @param isUsedForRootElement The isUsedForRootElement to set.
	 */
	public void setUsedForRootElement(boolean isUsedForRootElement) {

		this.isUsedForRootElement = isUsedForRootElement;
	}

	public boolean isFromSource() {

		if (binding != null)
			return binding.isFromSource();
		else
			return false;
	}

	public boolean isPublic() {

		if (binding != null)
			return Modifier.isPublic(getBinding().getModifiers());
		else
			return true;
	}
}