/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.ArrayAccess;
import org.eclipse.jdt.core.dom.ArrayCreation;
import org.eclipse.jdt.core.dom.ArrayInitializer;
import org.eclipse.jdt.core.dom.CastExpression;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.ConditionalExpression;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldAccess;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.ParenthesizedExpression;
import org.eclipse.jdt.core.dom.SuperFieldAccess;
import org.eclipse.jdt.core.dom.SuperMethodInvocation;
import org.eclipse.jdt.core.dom.ThisExpression;
import org.intoJ.inferType.calculation.InverseBindingResolver;
import org.intoJ.inferType.exception.AnonymousTypeException;
import org.intoJ.inferType.exception.IntoJCalculationException;
import org.intoJ.inferType.model.model.DeclarationElement;

/**
 * 
 * Contains utility functions for working with IBindings.
 * 
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * 
 * @author Andreas Mei&szlig;ner
 */
public class Util_IBinding {

	/**
	 * 
	 * This method is used to find invocations (method calls, field access) on
	 * "elements".
	 * 
	 * The method basically checks if the "access" performed in "theExpr" is
	 * being performed on an element represented by the binding "theBinding".
	 * 
	 * For example, "theBinding" may be a field called "field1". This method
	 * checks for any arbitrary expression, whether "field1" plays a part in the
	 * expression. It the expression is, say, "field1.a()", it plays a role and
	 * the method returns true. If however, for example, the expression is "new
	 * ClassA().doSomething()", then quite obviously "field1" has nothing to do
	 * with it.
	 * 
	 * This method investigates: 
	 * FieldAccess - this._a 
	 * SuperFieldAccess - super._a 
	 * ArrayAccess - _a[0] Name - _a || _o._a 
	 * MethodInvocation - this._a.method() 
	 * SuperMethodInvocation - super._a.method() 
	 * CastExpression - (Cast) _a 
	 * ConditionalExpression - (boolean) ? _a : _b
	 * ParenthesizedExpression - ((Cast) (Cast) _a) 
	 * ArrayInitializer - array = {_a, _b, _c}
	 * 
	 * @param theExpr
	 * @param theDE
	 * 
	 * @return true if given DE is inside the given Expression
	 */
	public static boolean isDEInsideExpression(Expression theExpr, DeclarationElement theDE) {

		// Kept for historic reasons. 
		boolean lookAlsoInsideCasts = false;

		if (theExpr == null)
			return false;
		if (theDE == null)
			return false;

		// get binding element of expression
		IBinding currentBinding = null;
		if (theExpr instanceof FieldAccess) // e.g.: this._a
			currentBinding = ((FieldAccess) theExpr).resolveFieldBinding();
		if (theExpr instanceof Name) // e.g.: _a || _o._a;
			currentBinding = ((Name) theExpr).resolveBinding();
		if (theExpr instanceof SuperFieldAccess) // e.g.: super._a
			currentBinding = ((SuperFieldAccess) theExpr).resolveFieldBinding();
		if (theExpr instanceof MethodInvocation) // e.g.: this._a.method()
			currentBinding = ((MethodInvocation) theExpr).resolveMethodBinding();
		if (theExpr instanceof SuperMethodInvocation) // e.g.: super._a.method()
			currentBinding = ((SuperMethodInvocation) theExpr).resolveMethodBinding();
		if (theExpr instanceof ConditionalExpression) // e.g.: (boolean) ? _a :
			// _b
			return (isDEInsideExpression(((ConditionalExpression) theExpr).getThenExpression(), theDE) || isDEInsideExpression(
					((ConditionalExpression) theExpr).getElseExpression(), theDE));
		if (theExpr instanceof ParenthesizedExpression) // e.g.: ((Cast) (Cast)
			// _a)
			return isDEInsideExpression(((ParenthesizedExpression) theExpr).getExpression(), theDE);

		// array stuff
		if (theExpr instanceof ArrayAccess) // e.g.: _a[0]
			return isDEInsideExpression(((ArrayAccess) theExpr).getArray(), theDE);
		if (theExpr instanceof ArrayInitializer) { // e.g.: array = {_a, _b, _c}
			ArrayInitializer arrayInitializer = (ArrayInitializer) theExpr;
			List expressions = arrayInitializer.expressions();
			Iterator expressionIterator = expressions.iterator();
			while (expressionIterator.hasNext())
				if (isDEInsideExpression((Expression) expressionIterator.next(), theDE))
					return true;
			return false;
		}
		//ARRAY: "Jump over" array creations
		if (theExpr instanceof ArrayCreation) // e.g.: new A[] { _a };
			return isDEInsideExpression(((ArrayCreation) theExpr).getInitializer(), theDE);

		// casts
		if (theExpr instanceof CastExpression) // e.g.: (Cast) _a
			return lookAlsoInsideCasts && isDEInsideExpression(((CastExpression) theExpr).getExpression(), theDE);

		// type stuff -- this will not be necessary
		//
		// the following handle expressions involving "this" or class creation
		// expressions.
		// unclear what to return here: constructors or the type itself?
		if (theExpr instanceof ClassInstanceCreation) // e.g.: new ClassA().a();
			// -> return type binding.
			currentBinding = ((ClassInstanceCreation) theExpr).resolveTypeBinding();
		if (theExpr instanceof ThisExpression) // e.g.: this.a(); -> return type
			// binding
			currentBinding = ((ThisExpression) theExpr).resolveTypeBinding();

		// if none of the previously listed nodes was visited return false
		if (currentBinding == null)
			return false;

		return (theDE.hasKey(currentBinding.getKey()));
	}

	/**
	 * This method finds all IBindings inside an Expression.
	 * 
	 * Note that bindingList will be filled recursively.
	 * 
	 * @param theExpr
	 *            the Expression
	 * @param bindingList
	 *            List of IBindings to be filled
	 */
	public static void getBindings(Expression theExpr, List bindingList) {

		if (theExpr instanceof FieldAccess) // e.g.: this._a
			bindingList.add(((FieldAccess) theExpr).resolveFieldBinding());
		if (theExpr instanceof Name) // e.g.: _a || _o._a;
			bindingList.add(((Name) theExpr).resolveBinding());
		if (theExpr instanceof SuperFieldAccess) // e.g.: super._a
			bindingList.add(((SuperFieldAccess) theExpr).resolveFieldBinding());
		if (theExpr instanceof ConditionalExpression) { // e.g.: (boolean) ? _a
			// : _b
			getBindings(((ConditionalExpression) theExpr).getThenExpression(), bindingList);
			getBindings(((ConditionalExpression) theExpr).getElseExpression(), bindingList);
		}
		if (theExpr instanceof ParenthesizedExpression) // e.g.: ((Cast) (Cast)
			// _a)
			getBindings(((ParenthesizedExpression) theExpr).getExpression(), bindingList);
		if (theExpr instanceof MethodInvocation) // e.g.: method()
			bindingList.add(((MethodInvocation) theExpr).resolveMethodBinding());
		if (theExpr instanceof SuperMethodInvocation) // e.g.: super.method()
			bindingList.add(((SuperMethodInvocation) theExpr).resolveMethodBinding());
		if (theExpr instanceof CastExpression) {// e.g.: (Cast) _a
			getBindings(((CastExpression) theExpr).getExpression(), bindingList);
		}
		// add typeBinding
		if (theExpr instanceof ClassInstanceCreation) // e.g.: new ClassA()
			bindingList.add(((ClassInstanceCreation) theExpr).resolveTypeBinding());
		if (theExpr instanceof ThisExpression)
			bindingList.add(((ThisExpression) theExpr).resolveTypeBinding());
		// array stuff
		if (theExpr instanceof ArrayAccess) // e.g.: _a[0]
			getBindings(((ArrayAccess) theExpr).getArray(), bindingList);
		if (theExpr instanceof ArrayCreation) // e.g.: new ClassA[5]
			bindingList.add(((ArrayCreation) theExpr).resolveTypeBinding());
		if (theExpr instanceof ArrayInitializer) { // e.g.: array = {_a, _b, _c}
			ArrayInitializer arrayInitializer = (ArrayInitializer) theExpr;
			List expressions = arrayInitializer.expressions();
			Iterator expressionIterator = expressions.iterator();
			while (expressionIterator.hasNext())
				getBindings((Expression) expressionIterator.next(), bindingList);
		}
	}

	/**
	 * This methods returns the parameterlist (List of
	 * SingleVariableDeclarations) of a given IMethodBinding.
	 * 
	 * @param binding the IMethodBinding
	 * @param javaProject the current IJavaProject
	 * 
	 * @return parameterlist (List of SingleVariableDeclarations)
	 * 
	 * @throws IntoJCalculationException
	 *             is thrown if it isnt possible to get the compilation unit of
	 *             the declaring type of the method (f.e. if the type is a
	 *             binary type)
	 * @throws AnonymousTypeException
	 */
	public static List getParametersOfMethod(IMethodBinding binding) throws IntoJCalculationException, AnonymousTypeException {

		List parameters = new ArrayList();

		// findType() does not find secondary types.
		IType declaringTypeOfParam = InverseBindingResolver.getITypeForITypeBinding(binding.getDeclaringClass());

		ICompilationUnit icu = declaringTypeOfParam.getCompilationUnit();
		if (icu == null)
			throw new IntoJCalculationException("", "");
		Util_ITypes u = new Util_ITypes();
		CompilationUnit cu = u.parseCompilationUnit(icu, true);

		MethodDeclarationVisitor methodDeclarationVisitor = new MethodDeclarationVisitor(binding);
		cu.accept(methodDeclarationVisitor);
		parameters = methodDeclarationVisitor.getParameterList();

		return parameters;
	}

	/**
	 * This visitor finds the parameterlist (List of SingleVariableDeclarations)
	 * of a given IMethodBinding.
	 * 
	 * @author Andreas Mei&szlig;ner
	 */
	private static class MethodDeclarationVisitor extends ASTVisitor {

		private IMethodBinding methodBinding;

		private List parameterList;

		/**
		 * Constructor for MethodDeclarationVisitor.
		 * @param methodBinding IMethodBinding
		 */
		public MethodDeclarationVisitor(IMethodBinding methodBinding) {

			this.methodBinding = methodBinding;
		}

		/**
		 * Method visit.
		 * @param node MethodDeclaration
		 * @return boolean
		 */
		public boolean visit(MethodDeclaration node) {

			if (node == null)
				return true;
			if (!methodBinding.getKey().equals(node.resolveBinding().getKey()))
				return true;

			parameterList = node.parameters();
			return true;
		}

		/**
		 * Method getParameterList.
		 * @return List
		 */
		public List getParameterList() {

			return parameterList;
		}
	}

}