/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.creation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.ThisExpression;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.internal.corext.dom.Bindings;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ParameterTypeVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ReturnTypeVariable2;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.TAA3EnhancedConstraintDataModel;
import org.intoJ.inferType3.internal.constraintModel.variables.ElementVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.GenericMethodTypeArgumentVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutabilityReason;
import org.intoJ.inferType3.internal.constraintModel.variables.IndependentReturnTypeVariable;
import org.intoJ.inferType3.internal.util.BindingUtil;
import org.intoJ.inferType3.internal.util.Cache;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * Handles the build up of the constraint model for everything
 * related to method invocations and declarations. Outsourced from
 * {@link InferTypeConstraintsCreator} for clarity.
 * 
 * @author Hannes Kegel
 *
 */
public class MethodHandler {
    private TypeDeclElementHandler typeDeclElementHandler;
    private InferTypeConstraintsModel model;
    private RefactoringStatus status;
    
    /**
     * The single constructor.
     * 
     * @param typeDeclElementHandler the handler for parametrized supertypes. Needed by this class
     * to register related method invocations and declarations.
     * @param model the model to be filled
     * @param status the current refactoring status
     */
    public MethodHandler(TypeDeclElementHandler typeDeclElementHandler, InferTypeConstraintsModel model, RefactoringStatus status) {
        this.typeDeclElementHandler = typeDeclElementHandler;
        this.model = model;
        this.status = status;    
    }
    
    /**
     * Handle a method invocation node. The parameters will be connected with the arguments,
     * the return type will be bound to the node and the method declaration requirement will
     * be bound to the expression.  
     * 
     * @param node the ASTNode of the invocation
     * @param currentTypeStack the stack of the declared types. Used for inner type handling.
     */
    public void handleInvocation(final MethodInvocation node, Stack<ITypeBinding> currentTypeStack, boolean includeThisReceiver) {
        final IMethodBinding binding = node.resolveMethodBinding();
        final Expression expression = node.getExpression();
        
        if (binding != null) {            
            if ("getClass".equals(binding.getName()) && binding.getParameterTypes().length == 0) {
                // special case for getClass wildcard
                handleGetClass(node, binding, expression);                
            } else {
                Map<String, GenericMethodTypeArgumentVariable> methodTypeVariables= createMethodTypeArguments(binding, expression);
                ConstraintVariable2 retTypeVar = handleInvocationReturn(expression, binding, methodTypeVariables);
                if (retTypeVar != null) {
                    InferTypeConstraintsCreator.setConstraintVariable(node, retTypeVar);
                }
                handleInvocationParameters(node.arguments(), binding, expression, methodTypeVariables, currentTypeStack, includeThisReceiver);
                ConstraintVariable2 descendant = null;
                if (expression != null && !(expression instanceof ThisExpression)) {
                	descendant = InferTypeConstraintsCreator.getConstraintVariable(expression);
                } else if (includeThisReceiver) {
                    if (expression instanceof ThisExpression) {
                        ThisExpression thisExpression = (ThisExpression)expression;
                        ITypeBinding calledOnType = thisExpression.resolveTypeBinding();
                        descendant = model.makeThisVariable(calledOnType);
                    } else {
	                    ITypeBinding declaringClass = binding.getDeclaringClass();
	
	                    for (int i=currentTypeStack.size() - 1; i >=0; i--) {
	                        ITypeBinding calledOnType = currentTypeStack.get(i); 
	                        if (BindingUtil.isFirstSubclassOrEqual(calledOnType, declaringClass)) {
	                            descendant = model.makeThisVariable(calledOnType);
	                            break;
	                        }
	                    }
                    }
                }
                if (expression != null && descendant == null && Modifier.isStatic(binding.getModifiers()) && includeThisReceiver) {
                	ITypeBinding staticBinding = expression.resolveTypeBinding();
                    descendant = model.makeStaticVariable(staticBinding);
                }
                if (descendant != null) {
                    handleDeclarationRequirements(binding, descendant);
                }
            }
        }
    }

    /**
     * Handle all parameters of an invocation. This is also used by ConstrutorInvocation etc.
     * 
     * @param arguments the arguments (element type: <code>Expression</code>)
     * @param binding the method binding
     * @param receiver the receiver of the method
     * @param methodTypeVariables the type variables of a generic method. Can be null.
     * @param currentTypeStack the stack of the declared types. Used for inner type handling.
     * @param includeThisReceiver 
     */
    public void handleInvocationParameters(final List arguments, final IMethodBinding binding, Expression receiver, Map<String, GenericMethodTypeArgumentVariable> methodTypeVariables, Stack<ITypeBinding> currentTypeStack, boolean includeThisReceiver) {
        if (methodTypeVariables == null) {
            methodTypeVariables = Collections.emptyMap();
        }
        ITypeBinding declaringClass = binding.getDeclaringClass();
        IMethodBinding methodDeclaration = declaringClass.isRawType() ? binding : binding.getMethodDeclaration();
        ITypeBinding[] parameterTypes = methodDeclaration.getParameterTypes();
        ITypeBinding[] currentParameterTypes = binding.getParameterTypes();
        
        //LogUtil.debugOut2(this, "receiver for " + binding + ": " + receiver);
        for (int index = 0; index < arguments.size(); index++) {
            Expression expression = (Expression) arguments.get(index);
            ConstraintVariable2 descendant = InferTypeConstraintsCreator.getConstraintVariable(expression);
            
            ConstraintVariable2 ancestor = null;

            int declIndex = index;
            if (binding.isVarargs() && index >= parameterTypes.length) {
                declIndex = parameterTypes.length - 1;
            }
            ITypeBinding parameterType = parameterTypes[declIndex];
            if (parameterType.isArray()) {
                parameterType = parameterType.getElementType();
            }
            
            if ((receiver != null && !(receiver instanceof ThisExpression)) || methodTypeVariables.size() > 0) {
                if (parameterType.isTypeVariable()) {
                    ancestor = methodTypeVariables.get(parameterType.getKey());
                    if (ancestor == null) {
                        ConstraintVariable2 createdTypeCv = InferTypeConstraintsCreator.getConstraintVariable(receiver);
                        if (createdTypeCv == null) {
                            //LogUtil.debugOut2(this, "no cv for " + receiver);
                        } else {
                            //LogUtil.debugOut2(this, "HAS A cv for " + receiver);
                            ancestor = model.makeElementVariable(createdTypeCv, parameterType, model.createTType(currentParameterTypes[declIndex]), true);
                        }
                    }
                } else if (parameterType.isParameterizedType()) {
                    ConstraintVariable2 createdTypeCv = InferTypeConstraintsCreator.getConstraintVariable(receiver);
                    // Elem[retVal] =^= Elem[receiver]
                    TType declaredTType= model.createTType(parameterType);
                    ancestor = makeInvocationParameterTypeVariable(binding, declIndex);
                    model.createTypeVariablesEqualityConstraints(createdTypeCv, methodTypeVariables, ancestor, declaredTType, model.createTType(currentParameterTypes[declIndex]));
                }
            } else if (declaringClass.isParameterizedType()) {
                if (receiver == null) {
                    for (int i=currentTypeStack.size() - 1; i >=0; i--) {
                        ITypeBinding calledOnType = currentTypeStack.get(i); 
                        if (BindingUtil.isFirstSubclassOrEqual(calledOnType, declaringClass)) {
                            ancestor = makeParamterizedMethodInvocationVariable(binding, declIndex, calledOnType);
                            break;
                        }
                    }
                } else if (receiver instanceof ThisExpression) {
                    ThisExpression thisExpression = (ThisExpression)receiver;
                    ITypeBinding calledOnType = thisExpression.resolveTypeBinding();
                    ancestor = makeParamterizedMethodInvocationVariable(binding, declIndex, calledOnType);
                }
            }

            if (descendant != null) {
                if (ancestor != null) {
                    model.createSubtypeConstraint(descendant, ancestor);
                } 
                ConstraintVariable2 declaredAncestor = makeDeclarationParameterTypeVariable(methodDeclaration, declIndex);
                if (declaredAncestor != null && (ancestor == null || !declaredAncestor.equals(ancestor))) {
                    model.createSubtypeConstraint(descendant, declaredAncestor, false);
                }
            
                if (!declaringClass.isFromSource() && declaredAncestor != null) {
                    String description = "used in non-source method " + declaringClass.getQualifiedName() + "." + binding.getName();
                    model.createSubtypeConstraint(declaredAncestor, model.makeImmutableTypeVariable(model.createTType(parameterTypes[declIndex]), description, null, ImmutabilityReason.OTHER));
                    model.makeElementsImmutable(declaredAncestor, description);
                }
            }
        }
    }
    
    /**
     * Handle the declaration of a method. The parameter and return constraint variables will be 
     * created and the connection to the original methods will be made.
     * 
     * @param node the ASTNode of the declaration
     */
    public void handleMethodDeclaration(final MethodDeclaration node) {
        final IMethodBinding binding = node.resolveBinding();
        if (binding == null) {
            return;
        }

        
        Collection<IMethodBinding> originals = null;
        if (!binding.isConstructor()) {
            final Type type = node.getReturnType2();
            if (type != null) {
                final ConstraintVariable2 first = model.makeReturnTypeVariable(binding, false);
                final ConstraintVariable2 second = InferTypeConstraintsCreator.getConstraintVariable(type);
                if (first != null) {
                    if (second != null) {
                        model.createEqualityConstraint(first, second);
                    }
                    originals = getRelatedMethods(binding);
                    handleOriginalReturns(binding, originals, first);
                }
            }
        }

        final List parameters = node.parameters();
        if (!parameters.isEmpty()) {
            if (originals == null) {
                originals = getRelatedMethods(binding);
            }
            SingleVariableDeclaration declaration = null;
            for (int index = 0; index < parameters.size(); index++) {
                declaration = (SingleVariableDeclaration) parameters.get(index);
                ParameterTypeVariable2 declParamVar = makeDeclarationParameterTypeVariable(binding, index);
                if (declParamVar != null) {
                    InferTypeConstraintsModel.setSimpleName(declParamVar, declaration.getName().toString());
                    ConstraintVariable2 descendant = InferTypeConstraintsCreator.getConstraintVariable(declaration.getType());
                    if (descendant != null) {
                        model.createEqualityConstraint(descendant, declParamVar);
                    }
                    IVariableBinding variable = declaration.resolveBinding();
                    InferTypeConstraintsModel.setSimpleName(declParamVar, variable.getName());
                    if (variable != null) {
                        descendant = model.makeVariableVariable(variable);
                        if (descendant != null) {
                        	TAA3EnhancedConstraintDataModel.setCorrespondingParameterTypeVar(descendant, declParamVar);
                            model.createEqualityConstraint(declParamVar, descendant);
                        }
                    }
                    for (IMethodBinding original : originals) {
                        if (!original.getKey().equals(binding.getKey())) {
                            descendant = makeOriginalDeclarationParameterTypeVariable(original, index, binding, declParamVar);
                            if (descendant != null) {
                                InferTypeConstraintsModel.setSimpleName(descendant, variable.getName());
                            }
                            model.createEqualityConstraint(declParamVar, descendant);
                        }
                    }
                }
            }
        }
        final List exceptions = node.thrownExceptions();
        if (!exceptions.isEmpty()) {
            final ITypeBinding throwable = node.getAST().resolveWellKnownType("java.lang.Throwable"); //$NON-NLS-1$
            if (throwable != null) {
                ConstraintVariable2 ancestor = model.makeImmutableTypeVariable(throwable, null, "thrown as exception", null, ImmutabilityReason.OTHER);
                if (ancestor != null) {
                    Name exception = null;
                    for (int index = 0; index < exceptions.size(); index++) {
                        exception = (Name) exceptions.get(index);
                        ConstraintVariable2 descendant = InferTypeConstraintsCreator.getConstraintVariable(exception);
                        if (descendant != null) {
                            model.createSubtypeConstraint(descendant, ancestor);
                        }
                    }
                }
            }
        }
    }

    /**
     * Handle the return type of an invocation.
     * 
     * @param expression the receiver of the method
     * @param binding the method binding
     * @param methodTypeVariables the optional type variables of a generic method
     * @return an {@link ElementVariable}, an {@link IndependentReturnTypeVariable} or a {@link ReturnTypeVariable2}
     */
    public ConstraintVariable2 handleInvocationReturn(final Expression expression, final IMethodBinding binding, Map<String, GenericMethodTypeArgumentVariable> methodTypeVariables) {
        if (!binding.isConstructor()) { 
            if (methodTypeVariables == null) {
                methodTypeVariables = Collections.emptyMap();
            }
            IMethodBinding methodDeclaration = binding.getDeclaringClass().isRawType() ? binding : binding.getMethodDeclaration();
//            for (Map.Entry<String, GenericMethodTypeArgumentVariable> entry : methodTypeVariables.entrySet()) {
//                LogUtil.debugOut(this, "type variable: " + entry.getKey() + ": " + entry.getValue());
//            }
            
            //LogUtil.debugOut2(this, "expr " + expression + " for " + binding);
            if (expression != null || methodTypeVariables.size() > 0) {            
                ITypeBinding returnTypeBinding = methodDeclaration.getReturnType();
                if (returnTypeBinding.isArray()) {
                    returnTypeBinding = returnTypeBinding.getElementType();
                }
                ConstraintVariable2 expressionCv = InferTypeConstraintsCreator.getConstraintVariable(expression);
                if (returnTypeBinding.isTypeVariable()) {
                    ConstraintVariable2 methodTypeVar = methodTypeVariables.get(returnTypeBinding.getKey());
                    if (methodTypeVar != null) {
                        return methodTypeVar;
                    }
                    if (expressionCv == null) {
                        //LogUtil.debugOut2(this, "no ret cv for " + expression);
                    } else {
                        return model.makeElementVariable(expressionCv, returnTypeBinding, model.createTType(binding.getReturnType()), true); 
                        //LogUtil.debugOut2(this, "HAS A ret cv for " + expression + ", variable: " + variable);
                    }
                } else if (returnTypeBinding.isParameterizedType()) {
                    TType declaredReturnTType= model.createTType(returnTypeBinding);
                    ConstraintVariable2 variable = model.makeReturnTypeVariable(
                            binding, 
                            containsWildcardsOrTypeVars(returnTypeBinding));                        
                    model.createTypeVariablesEqualityConstraints(expressionCv, methodTypeVariables, variable, declaredReturnTType, model.createTType(binding.getReturnType()));
                    return variable;
                } else {           
                    return model.makeReturnTypeVariable(binding, false);
                }
            } else {           
                return model.makeReturnTypeVariable(binding, false);
            }
        }
        return null;
    }        
    
    public void handleDeclarationRequirements(final IMethodBinding binding, final ConstraintVariable2 descendant) {
        //LogUtil.debugOut2(this, "end visit: " + binding + ", originals: " + originals.size());
        for (IMethodBinding originalMethod : getOriginalMethods(binding)) {
            ITypeBinding originalDeclaringClass = originalMethod.getDeclaringClass();
            if (originalDeclaringClass != null) {
                final ConstraintVariable2 ancestor;
                int modifiers = originalMethod.getModifiers();
                if (Modifier.isStatic(modifiers)) {
                	EnumSet<ImmutabilityReason> reason; 
                	if (Modifier.isProtected(modifiers)) {
                		reason = EnumSet.of(ImmutabilityReason.STATIC_METHOD, ImmutabilityReason.PROTECTED_METHOD);
                	} else if (!Modifier.isPublic(modifiers)) {
                		reason = EnumSet.of(ImmutabilityReason.STATIC_METHOD, ImmutabilityReason.PACKAGE_METHOD);
                	} else {
                		reason = EnumSet.of(ImmutabilityReason.STATIC_METHOD);
                	}
                    ancestor = model.makeImmutableDeclaringTypeVariable(originalDeclaringClass, "access to static method " + binding.getDeclaringClass().getQualifiedName() + "." + binding.getName(), originalMethod.getMethodDeclaration().getKey(), reason);
                } else if (Modifier.isProtected(modifiers)) {
                    ancestor = model.makeImmutableDeclaringTypeVariable(originalDeclaringClass, "access to protected method " + binding.getDeclaringClass().getQualifiedName() + "." + binding.getName(), originalMethod.getMethodDeclaration().getKey(), ImmutabilityReason.PROTECTED_METHOD);
                } else if (!Modifier.isPublic(modifiers)) {
                    ancestor = model.makeImmutableDeclaringTypeVariable(originalDeclaringClass, "access to package-private method " + binding.getDeclaringClass().getQualifiedName() + "." + binding.getName(), originalMethod.getMethodDeclaration().getKey(), ImmutabilityReason.PACKAGE_METHOD);
                } else {
                    ancestor = model.makeDeclaringTypeVariable(originalMethod, originalDeclaringClass);
                }
                if (ancestor != null)
                    model.createSubtypeConstraint(descendant, ancestor);
            }
        }
    }
    
    private void handleOriginalReturns(final IMethodBinding binding, Collection<IMethodBinding> originals, ConstraintVariable2 declVar) {
        if (declVar != null) {
            for (IMethodBinding originalBinding : originals) {
                if (!originalBinding.getKey().equals(binding.getKey())) {
                    ConstraintVariable2 ancestor = makeOriginalReturnTypeVariable(originalBinding, binding, declVar);
                    if (ancestor != null) {
                        model.createCovariantTypeConstraint(declVar, binding.getReturnType(), ancestor, originalBinding.getReturnType());
                    }
                }                
            }
        }
    }

    private void handleGetClass(final MethodInvocation node, final IMethodBinding binding, final Expression expression) {
        //special case: see JLS3 4.3.2
        ITypeBinding returnType = binding.getReturnType();
        
        if (returnType.isParameterizedType() && expression != null) {
            String classKey = returnType.getTypeDeclaration().getTypeParameters()[0].getKey();
            
            final ConstraintVariable2 expressionCv = InferTypeConstraintsCreator.getConstraintVariable(expression);
            if (expressionCv != null) {
                ConstraintVariable2 retTypeVar = model.makeReturnTypeVariable(binding, true);
                ElementVariable elementVariable = InferTypeConstraintsModel.getElementVariables(retTypeVar).get(classKey);
                
                if (retTypeVar != null && elementVariable != null) {
                    InferTypeConstraintsCreator.setConstraintVariable(node, retTypeVar);
                    model.createSubtypeConstraint(expressionCv, elementVariable);
                }                                                
            }
        }
    }

    private Map<String, GenericMethodTypeArgumentVariable> createMethodTypeArguments(IMethodBinding methodBinding, Expression expression) {
        ITypeBinding[] methodTypeParameters= methodBinding.getMethodDeclaration().getTypeParameters();
        ITypeBinding[] methodTypeArguments = methodBinding.getTypeArguments();
        Map<String, GenericMethodTypeArgumentVariable> methodTypeVariables;
        if (methodTypeParameters.length == 0) {
            methodTypeVariables= Collections.emptyMap();
        } else {
            methodTypeVariables= new HashMap<String, GenericMethodTypeArgumentVariable>();
            for (int i= 0; i < methodTypeParameters.length; i++) {
                ITypeBinding methodTypeParameter= methodTypeParameters[i];
                
                TType argument = model.createTType(methodTypeArguments[i]);
                GenericMethodTypeArgumentVariable typeVariableCv= model.makeGenericMethodTypeArgumentVariable(argument, 
                        methodTypeVariables, 
                        InferTypeConstraintsCreator.getConstraintVariable(expression), 
                        methodTypeParameter);
                methodTypeVariables.put(methodTypeParameter.getKey(), typeVariableCv);
            }            
        }

        return methodTypeVariables;
        
    }
    
    private Collection<IMethodBinding> getOriginalMethods(IMethodBinding methodBinding) {
        /* The search for original methods is different from "Use Supertype Where Possible"
         * to correctly handle a "marriage problem" when a type inherits a method from two 
         * different supertypes, but does not declare it. The handling here is similar to the 
         * RippleMethodFinder2.
         */
        final Collection<IMethodBinding> originals = new HashSet<IMethodBinding>();

        if (methodBinding.isConstructor() || Modifier.isPrivate(methodBinding.getModifiers())
                || Modifier.isStatic(methodBinding.getModifiers())) 
        {
            originals.add(methodBinding);
        } else {
            ITypeBinding declaringClass = methodBinding.getDeclaringClass();
            ITypeBinding[] allSuperTypes= Bindings.getAllSuperTypes(declaringClass);
    
            for (int i= 0; i < allSuperTypes.length; i++) {
                ITypeBinding superType= allSuperTypes[i];
                IMethodBinding superMethod= Bindings.findOverriddenMethodInType(superType, methodBinding);
                if (superMethod != null) {
                    originals.add(superMethod);
                }
            }
            originals.add(methodBinding);
        }
            
        return originals;       
    }
    
    private static List<IType> getSubtypeLeafsWithInferfaces(ITypeHierarchy hierarchy, IType baseType, boolean parentInterfaces) throws JavaModelException {
        List<IType> list = new ArrayList<IType>();
        
        IType[] subTypes = hierarchy.getSubtypes(baseType);
        for (IType type : subTypes) {
        	boolean hasInterfaces = parentInterfaces ? true : type.getSuperInterfaceNames().length > 0;        	
            List<IType> subSubLeaves = getSubtypeLeafsWithInferfaces(hierarchy, type, hasInterfaces);
            if (subSubLeaves.isEmpty()) {
            	if (hasInterfaces) {
                    list.add(type);            		
            	}
            } else {
                list.addAll(subSubLeaves);
            }
        }
        
        return list;        
    }
    
    private Collection<IMethodBinding> getRelatedMethods(IMethodBinding methodBinding) {
        /* The search for original methods is different from "Use Supertype Where Possible"
         * to correctly handle a "marriage problem" when a type inherits a method from two 
         * different supertypes, but does not declare it. The handling here is similar to the 
         * RippleMethodFinder2.
         */
        final Collection<IMethodBinding> originals = new ArrayList<IMethodBinding>();

        if (methodBinding.isConstructor() || Modifier.isPrivate(methodBinding.getModifiers())
                || Modifier.isStatic(methodBinding.getModifiers())) 
        {
            return originals;
        }
            
        Set<String> handledSuperTypes = new HashSet<String>(); 
        ITypeBinding declaringClass = methodBinding.getDeclaringClass();
        ITypeBinding[] allSuperTypes= Bindings.getAllSuperTypes(declaringClass);
    
        for (int i= 0; i < allSuperTypes.length; i++) {
            ITypeBinding superType= allSuperTypes[i];
            handledSuperTypes.add(superType.getQualifiedName());
            IMethodBinding superMethod= Bindings.findOverriddenMethodInType(superType, methodBinding);
            if (superMethod != null) {
                originals.add(superMethod);
            }
        }
        handledSuperTypes.add(declaringClass.getQualifiedName());
    
        if (Modifier.isPublic(methodBinding.getModifiers()) && 
                !declaringClass.isInterface() && declaringClass.isFromSource() && 
                !declaringClass.getQualifiedName().equals("java.lang.Object")) 
        {      
            try {
                IMethod searchedMethod = (IMethod) methodBinding.getJavaElement();
                
                IType declaringType = (IType)declaringClass.getJavaElement();
                ITypeHierarchy subHierarchy = Cache.getHierarchy(null, declaringType, true);
                if (subHierarchy != null) {                    
                    List<IType> subTypes = getSubtypeLeafsWithInferfaces(subHierarchy, declaringType, false);
                    for (IType subType : subTypes) {
                        ITypeHierarchy superHierarchy = Cache.getSuperHierarchy(subType);
                        IType[] superTypes = superHierarchy.getAllSupertypes(subType);
                        for (int superIndex = 0; superIndex < superTypes.length; superIndex++) {
                            if (superTypes[superIndex].isInterface() && handledSuperTypes.add(superTypes[superIndex].getFullyQualifiedName('.'))) {
                                //check IMethods first for performance reasons
                                IMethod[] superMethods = superTypes[superIndex].getMethods();                            
                                for (int i = 0; i < superMethods.length; i++) {
                                    if (superMethods[i].isSimilar(searchedMethod)) {
                                        ITypeBinding superBinding = Cache.getBinding(superTypes[superIndex]);
                                        IMethodBinding superMethod= Bindings.findOverriddenMethodInType(superBinding, methodBinding);
                                        if (superMethod != null) {
                                            originals.add(superMethod);
                                        }
                                        break;
                                    }
                                }
                                
                            }
                        }
                    }
                    if (subTypes.isEmpty()) {
                        Cache.markHierarchyAsUnused(declaringType);
                    }
                    
                }
                
            } catch (InternalCalculationException e) {
                // ok, type cannot be parsed
            } catch (CoreException e) {
                e.printStackTrace();
                status.addFatalError(e.toString());
            }
        }
    
            
        return originals;
    }
    
    private ConstraintVariable2 makeOriginalReturnTypeVariable(IMethodBinding originalBinding, IMethodBinding declBinding, ConstraintVariable2 declVar) {
        IMethodBinding methodDecl = originalBinding.getMethodDeclaration();
        ConstraintVariable2 origVar = null;
        if (originalBinding.getDeclaringClass().isFromSource()) {
            origVar = model.makeReturnTypeVariable(methodDecl, false);
        } else {
            origVar = model.makeImmutableDeclaringTypeVariable(originalBinding.getMethodDeclaration().getReturnType(), 
                    "Binary method " + originalBinding.getMethodDeclaration() + " would have to be changed.", null, ImmutabilityReason.OTHER);
        }
        if (origVar != null) {
            TType origType = model.createTType(originalBinding.getDeclaringClass());
            TType methodDeclType = model.createTType(declBinding.getDeclaringClass());
            ITypeBinding parameterBinding = methodDecl.getReturnType();
            typeDeclElementHandler.registerParameter(originalBinding, parameterBinding, origType, methodDeclType, declVar, model);
        }
        return origVar;
    }

    private ConstraintVariable2 makeOriginalDeclarationParameterTypeVariable(IMethodBinding originalBinding, int index, IMethodBinding declBinding, ParameterTypeVariable2 declVar) {
        IMethodBinding methodDecl = originalBinding.getMethodDeclaration();
        ConstraintVariable2 origVar = null;
        if (originalBinding.getDeclaringClass().isFromSource()) {
            origVar = model.makeParameterTypeVariable(methodDecl, index, false);
        } else {
            origVar = model.makeImmutableDeclaringTypeVariable(methodDecl.getParameterTypes()[index], 
                    "Binary method " + originalBinding.getMethodDeclaration() + " would have to be changed.", null, ImmutabilityReason.OTHER);
        }
        if (origVar != null) {
            TType origType = model.createTType(originalBinding.getDeclaringClass());
            TType methodDeclType = model.createTType(declBinding.getDeclaringClass());
            ITypeBinding parameterBinding = methodDecl.getParameterTypes()[declVar.getParameterIndex()];

            typeDeclElementHandler.registerParameter(originalBinding, parameterBinding, origType, methodDeclType, declVar, model);
        }
        return origVar;
    }
    
    private ConstraintVariable2 makeParamterizedMethodInvocationVariable(IMethodBinding originalBinding, int index, ITypeBinding usedOnType) {
        IMethodBinding methodDecl = originalBinding.getMethodDeclaration();

        ConstraintVariable2 uniqueVar = model.makeParameterTypeVariable(originalBinding, index, mustBeIndependent(originalBinding, index));
        if (uniqueVar != null) {
            TType origType = model.createTType(originalBinding.getDeclaringClass());
            TType methodUsedType = model.createTType(usedOnType);
            ITypeBinding parameterBinding = methodDecl.getParameterTypes()[index];
    
            typeDeclElementHandler.registerParameter(originalBinding, parameterBinding, origType, methodUsedType, uniqueVar, model);
        }
        return uniqueVar;
    }
    
    
    private ParameterTypeVariable2 makeDeclarationParameterTypeVariable(IMethodBinding methodBinding, int parameterIndex) {
        return (ParameterTypeVariable2)model.makeParameterTypeVariable(methodBinding.getMethodDeclaration(), parameterIndex, false);
    }

    private ConstraintVariable2 makeInvocationParameterTypeVariable(IMethodBinding methodBinding, int parameterIndex) {
        return model.makeParameterTypeVariable(methodBinding, parameterIndex, mustBeIndependent(methodBinding, parameterIndex));
    }
    
    private static boolean mustBeIndependent(IMethodBinding methodBinding, int parameterIndex) {
        return methodBinding.isParameterizedMethod() || containsWildcardsOrTypeVars(methodBinding.getMethodDeclaration().getParameterTypes()[parameterIndex]);
    }

    /**
     * Checks if the given binding contains a wildcard or a type variable.
     * @param typeBinding the binding to check
     * @return boolean
     */
    public static boolean containsWildcardsOrTypeVars(ITypeBinding typeBinding) {
        if (typeBinding.isArray()) {
            typeBinding = typeBinding.getElementType();
        }
        if (typeBinding.isWildcardType()) {
            return true;
        } else if (typeBinding.isTypeVariable()) {
            return true;
        } else if (typeBinding.isParameterizedType()) {
            for (ITypeBinding arg : typeBinding.getTypeArguments()) {
                if (containsWildcardsOrTypeVars(arg)) {
                    return true;
                }
            }
        }
        return false;
    }        
}
