/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.manipulation;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.jdt.core.IBuffer;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.ISourceRange;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;

/**
 * The utiliy class for the code generation.
 * 
 * @author Andreas Mei&szlig;ner
 * @author Hannes Kegel
 */
public class GenerationUtil {

	/**
	 * Creates an empty compilation unit and its correspondent packages from a
	 * fully qualified name.
	 * 
	 * @param fqName full qualified class name
	 * @param javaProject current IJavaProject
	 * @return empty java file compilation unit
	 * 
	 * @throws JavaModelException
	 * @throws NoSourceFolderException
	 * @throws CoreException
	 */
	public static ICompilationUnit createCompilationUnit(String fqName, IJavaProject javaProject) throws JavaModelException {

		String typeName = getSimpleTypeName(fqName);
		String packageName = getPackageName(fqName);

		IClasspathEntry[] classpathEntries = javaProject.getRawClasspath();
		IPath sourcePath = null;
		for (IClasspathEntry element : classpathEntries) {
			int entryKind = element.getEntryKind();
			if (entryKind == IClasspathEntry.CPE_SOURCE) {
				sourcePath = element.getPath();
				break;
			}
		}

		if (sourcePath == null) {
            throw new RuntimeException("no source folder");
        }

		IPackageFragmentRoot sourceFolder = javaProject.findPackageFragmentRoot(sourcePath);
		IPackageFragment packageFragment = sourceFolder.createPackageFragment(packageName, true, null);
		ICompilationUnit icu = packageFragment.getCompilationUnit(typeName + ".java");

		return icu;
	}

	/**
	 * Creates a StringBuffer with all required modifiers. Note that this
	 * StringBuffer ends with an blank.
	 * 
	 * @param methodBinding the binding
	 * @param createAbstractMethod boolean
	 * @return StringBuffer with modifiers
	 */
	public static StringBuffer getModifiers(IMethodBinding methodBinding, boolean createAbstractMethod) {

		int modifiers = methodBinding.getModifiers();
		return getModifiers(modifiers, createAbstractMethod);
	}

	public static StringBuffer getModifiers(int modifiers, boolean createAbstractMethod) {
		StringBuffer sBuf = new StringBuffer();
		if (Modifier.isPublic(modifiers)) {
            sBuf.append("public ");
        }
		if (Modifier.isProtected(modifiers)) {
            sBuf.append("protected ");
        }
		if (Modifier.isPrivate(modifiers)) {
            sBuf.append("private "); // cant happen
        }
		if (Modifier.isStatic(modifiers)) {
            sBuf.append("static "); 
        }

		if (createAbstractMethod) {
            sBuf.append("abstract ");
        }

		return sBuf;
	}

	/**
	 * Creates a StringBuffer with a parameter list. The parameter names are
	 * generated automatically (i.e. arg0, arg1, ..).
     * @param methodBinding the binding of the method 
	 * @param originalMethod the original IMethod used to get parameter names
     * @param importRewrite the import rewriter used to get simple names 
	 * @return StringBuffer with parameter list ("Type arg0,Type arg1,Type
	 *         arg2,..");
	 * @throws JavaModelException 
	 */
	public static StringBuffer getParametersList(IMethodBinding methodBinding, IMethod originalMethod, ImportRewrite importRewrite) throws JavaModelException {
        String[] parameterNames = originalMethod.getParameterNames();
        
		StringBuffer sBuf = new StringBuffer();

		ITypeBinding[] types = methodBinding.getParameterTypes();
		for (int i = 0; i < types.length; i++) {
            ITypeBinding typeBinding = types[i];
            String sep = " ";
            if (methodBinding.isVarargs() && i == types.length - 1) {
                typeBinding = typeBinding.getComponentType();
                sep = "... ";
            }
            
			sBuf.append(importRewrite.addImport(typeBinding)).append(sep).append(parameterNames[i]);
            if (i < types.length - 1) {
                sBuf.append(", ");
            }
		}

        return sBuf;
	}

	/**
	 * Creates a StringBuffer with a exceptions list. The several exceptions
	 * will be separated by comma.
     *
     * @param methodBinding the binding of the method 
     * @param importRewrite the import rewriter used to get simple names 
	 * @return StringBuffer with exceptions list ("exception1,exception2,..")
	 */
	public static StringBuffer getExceptionsList(IMethodBinding methodBinding, ImportRewrite importRewrite) {

		StringBuffer sBuf = new StringBuffer();

        ITypeBinding[] exceptionTypes = methodBinding.getExceptionTypes();
        for (int i = 0; i < exceptionTypes.length; i++) {
			sBuf.append(importRewrite.addImport(exceptionTypes[i]));
            if (i < exceptionTypes.length - 1) {
                sBuf.append(", ");
            }
		}
		return sBuf;
	}

    /**
     * Creates the type parameter list for the given method. Returns an empty buffer if this is
     * no generic method.
     *  
     * @param methodBinding the binding of the method 
     * @param importRewrite the import rewriter used to get simple names 
     * @return the type parameter list with bounds for generic methods, empty buffer otherwise 
     */
    public static StringBuffer getTypeParameterList(IMethodBinding methodBinding, ImportRewrite importRewrite) {
        return getTypeParameterList(methodBinding.getTypeParameters(), importRewrite);
    }

	public static StringBuffer getTypeParameterList(
			ITypeBinding[] typeParameters, ImportRewrite importRewrite) {
		StringBuffer sBuf = new StringBuffer();

        for (int i = 0; i < typeParameters.length; i++) {
            sBuf.append(importRewrite.addImport(typeParameters[i]));
            
            ITypeBinding[] bounds = typeParameters[i].getTypeBounds();
            for (int boundIndex = 0; boundIndex < bounds.length; boundIndex++) {
                if (boundIndex == 0) {
                    sBuf.append(" extends ");
                }
                sBuf.append(importRewrite.addImport(bounds[boundIndex]));

                if (boundIndex < bounds.length - 1) {
                    sBuf.append(" & ");
                }                
            }
            
            
            if (i < typeParameters.length - 1) {
                sBuf.append(", ");
            }
        }
        return sBuf;
	}

    /**
     * Retrieves the javadoc for the given method.
     * 
     * @param method the method
     * @return the javadoc or an empty string if there is none
     * @throws JavaModelException
     */
    public static String getJavadoc(IMethod method) throws JavaModelException {
        ISourceRange range = method.getJavadocRange();
        ICompilationUnit compilationUnit = method.getCompilationUnit();
        if (range != null && compilationUnit != null) {
            IBuffer buffer = compilationUnit.getBuffer();
            if (buffer != null) {
                return buffer.getText(range.getOffset(), range.getLength());
            }
        }
        return "";
    }
    
    /**
     * Convenience method for determining the type name.
     * 
     * @param fullyQualifiedTypeName
     * @return String
     */
    private static String getSimpleTypeName(String fullyQualifiedTypeName) {

        return fullyQualifiedTypeName.substring(fullyQualifiedTypeName.lastIndexOf('.') + 1);
    }

    /**
     * Convenience method for determining the package name.
     * 
     * @param fullyQualifiedTypeName
     * @return String
     */
    private static String getPackageName(String fullyQualifiedTypeName) {

        if (fullyQualifiedTypeName.lastIndexOf('.') < 0) {
            return "";
        }
        return fullyQualifiedTypeName.substring(0, fullyQualifiedTypeName.lastIndexOf('.'));
    }

    
}