/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc;

import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;

import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.ArrayType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.NullType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ReturnTypeVariable2;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.internal.constraintModel.EqualityTypeConstraint;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.util.ConstraintTraversal;
import org.intoJ.inferType3.internal.constraintModel.variables.BindingKeyProvider;
import org.intoJ.inferType3.internal.constraintModel.variables.DeclaringTypeVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ElementVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.GenericMethodTypeArgumentVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.IImmutableVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutableTypeVariableWithDescr;
import org.intoJ.inferType3.internal.constraintModel.variables.IndependentReturnTypeVariable;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;

/**
 * This class calculates the access sets for the selected variables based on the
 * constraints from the {@link InferTypeConstraintsModel}. 
 * <p>
 * Instances of {@link DeclaringTypeVariable} will get an initial access set with a singleton
 * of the invoked method, immutable variables get an immutable access set with a description.
 * </p>
 * <p>
 * In a depth-first post-order run the spanning tree of the component of the
 * constraint graph with the selected element as root is visited and the access sets
 * are merged back to the root of the tree.
 * </p>
 * @author Hannes Kegel
 *
 */
public class AccessSetCalculation {
    
    /**
     * This calculates the access sets for the given model (one could say
     * it solves the model).
     * 
     * @param model the {@link InferTypeConstraintsModel}
     * @param typeHierarchy the modifiable hierarchy of the selected type
     * @param status the refactoring status used to store errors
     * @param ignoreSuperfluousMethods <code>true</code> id the new type should
     * be added even if the old one is already minimal, <code>false</code> if
     * not
     * @throws JavaModelException
     */
    public static AccessSet calculate(InferTypeConstraintsModel model,
    		TypeHierarchy typeHierarchy, RefactoringStatus status,
    		boolean ignoreSuperfluousMethods) throws JavaModelException {
        initAccessSets(model, typeHierarchy, false);
        return calculateAccessSets(model, typeHierarchy, status, ignoreSuperfluousMethods);      
    }

    /**
     * This methods calculates the access sets for all variables in the model with a type included in the given hierarchy. 
     * @param model the constraint model
     * @param typeHierarchy the type hierarchy
     * @param b 
     * @return the merged set for all DEs with types from the hierarchy
     * @throws JavaModelException
     */
    public static AccessSet calculateAllSets(InferTypeConstraintsModel model, TypeHierarchy typeHierarchy, boolean allMethods) throws JavaModelException {
        initAccessSets(model, typeHierarchy, allMethods);
        
        AccessSet mergedSet = new AccessSet();
        for (ConstraintVariable2 cv : model.getConstraintVariables()) {
        	
        	// for TAA3:
        	if (cv.getType() instanceof NullType)
        		continue;
        	
            if (typeHierarchy.getType(cv.getType()) != null) {
                // returns immediately if already determined
            	mergedSet.addSet(calculateAccessSet(cv, new HashSet<ITypeConstraint2>(), new LinkedHashSet<ConstraintVariable2>()));
            }
        }
        return mergedSet;
    }

    private static AccessSet calculateAccessSets(InferTypeConstraintsModel model, TypeHierarchy typeHierarchy, RefactoringStatus status, boolean ignoreSuperfluousMethods) throws JavaModelException {
        ConstraintVariable2 variable = model.getSelectedConstraintVariable();
        if (variable == null) {
            if (model.getSelectedElement().isTypeSelection()) {
                status.addFatalError("No declaration elements of this type found.");                
            } else {
                status.addFatalError("Could not find selected element.");
            }
            return null;
        }
        
        AccessSet accessSet = calculateAccessSet(variable, new HashSet<ITypeConstraint2>(), new LinkedHashSet<ConstraintVariable2>());

        status.merge(model.getAndClearRefactoringStatus());
        
        if (accessSet.isUnmodfiable()) {
            status.addFatalError("The selected declaration element is unmodifiable:\n\n" + accessSet.getShortUnmodifiableDescription());
            //LogUtil.debugOut2(AccessSetCalculation.class, accessSet.getUnmodifiableDescription());
            //return;
        }
        
        if (!status.hasFatalError() && !ignoreSuperfluousMethods && !hasSuperfluousMethods(typeHierarchy, accessSet)) {
            status.addFatalError("The selected type is already minimal.");
        }
        

        /*
        LogUtil.debugOut(AccessSetCalculation.class, "access set for " + variable);
        if (variable instanceof ITypeConstraintVariable) {
            LogUtil.debugOut(AccessSetCalculation.class, "cu: " +((ITypeConstraintVariable) variable).getCompilationUnit().getElementName());
            LogUtil.debugOut(AccessSetCalculation.class, "offset: " +((ITypeConstraintVariable) variable).getRange().getSourceRange().getOffset());
            LogUtil.debugOut(AccessSetCalculation.class, "length: " +((ITypeConstraintVariable) variable).getRange().getSourceRange().getLength());
        }
        LogUtil.debugOut(AccessSetCalculation.class, accessSet);            
        LogUtil.debugOut(AccessSetCalculation.class, "----------------------------");
        */
        InferTypeConstraintsModel.setAccessSet(variable, accessSet);
        return accessSet;
    }
    
    private static AccessSet calculateAccessSet(final ConstraintVariable2 variable, final Set<ITypeConstraint2> currentProcessSet, final LinkedHashSet<ConstraintVariable2> pathFromRoot) throws JavaModelException {
        AccessSet previousAccessSet = InferTypeConstraintsModel.getAccessSet(variable);
        if (previousAccessSet != null) {
            return previousAccessSet;
        }
        
        pathFromRoot.add(variable);
        
        //LogUtil.debugOut(AccessSetCalculation.class, "start set for " + variable + "(" + System.identityHashCode(variable) + "): " + accessSet);

        InferTypeConstraintsModel.setAccessSet(variable, new AccessSet(variable));
        
        final boolean fromReturnType = isFromReturnType(variable);
         
        ConstraintTraversal.visitSingleStep(currentProcessSet, variable, 
                new ConstraintTraversal.Visitor() 
        {
            public boolean visit(ConstraintVariable2 baseConstraintVariable, ConstraintVariable2 otherConstraintVariable, ITypeConstraint2 constraint) throws JavaModelException {
                //LogUtil.debugOut(this, "following constraint: " + constraint);
                if (fromReturnType && otherConstraintVariable instanceof GenericMethodTypeArgumentVariable) {
                    InferTypeConstraintsModel.getAccessSet(variable).addSet(new AccessSet("used as unmodifiable type argument of a generic method", variable));                        
                } else {
                    if (pathFromRoot.contains(otherConstraintVariable)) { // backwards edge
                        mergeSets(otherConstraintVariable, baseConstraintVariable, pathFromRoot);
                    } else {
                        AccessSet rightSet = calculateAccessSet(otherConstraintVariable, currentProcessSet, pathFromRoot);
                        AccessSet leftSet = InferTypeConstraintsModel.getAccessSet(variable);
                        if (constraint instanceof EqualityTypeConstraint) {
                            // an equality constraint is treated like a tree and a backwards edge.
                            // In fact, as an undirected edge in a directed graph it would be modelled
                            // with two edges so this is consistent
                            leftSet.mergeSet(rightSet);                            
                        } else {
                            leftSet.addSet(rightSet);
                        }
                    }
                }
                return true;
            }
            
        }); 
        //LogUtil.debugOut(AccessSetCalculation.class, "end set for " + variable + "(" + System.identityHashCode(variable) + "): " + accessSet);
        pathFromRoot.remove(variable);

        return InferTypeConstraintsModel.getAccessSet(variable);
    }
    
    private static void mergeSets(ConstraintVariable2 fromCv, ConstraintVariable2 toCv, LinkedHashSet<ConstraintVariable2> pathFromRoot) {
        boolean inCycle = false;
        AccessSet singleSet = null;
        for (ConstraintVariable2 cv : pathFromRoot) {
            if (cv == fromCv) {
                inCycle = true;
                singleSet = InferTypeConstraintsModel.getAccessSet(cv);
            } else if (inCycle) {
                singleSet.mergeSet(InferTypeConstraintsModel.getAccessSet(cv));
            }
            if (cv == toCv) {
                break;
            }            
        }
    }
    
    private static boolean isFromReturnType(ConstraintVariable2 variable) {
        if (variable instanceof ReturnTypeVariable2 || 
                variable instanceof IndependentReturnTypeVariable) 
        {
            return true;            
        } else if (variable instanceof ElementVariable) {
            return isFromReturnType(((ElementVariable)variable).getBaseConstraintVariable());
        }
        return false;
    }

    private static void initAccessSets(InferTypeConstraintsModel model, TypeHierarchy typeHierarchy, boolean allMethods) {        
        for (ConstraintVariable2 variable : model.getConstraintVariables()) {
            
            if (variable instanceof DeclaringTypeVariable || (allMethods && variable instanceof BindingKeyProvider)) {
                String bindingKey = ((BindingKeyProvider)variable).getBindingKey();
				if (bindingKey != null && !bindingKey.startsWith("Ljava/lang/Object;")) {
                    
                    BindingWrapper bindingWrapper = typeHierarchy.getMethod(bindingKey);
                    
                    AccessSet accessSet;
                    if (bindingWrapper != null) {
                        accessSet = new AccessSet(bindingWrapper, variable);                        
                    } else {
                        accessSet = new AccessSet("method not in hierarchy: " + bindingKey, variable);                                                
                    }
                    InferTypeConstraintsModel.setAccessSet(variable, accessSet);

                    continue;
                }
            }
            
            if (variable instanceof ImmutableTypeVariableWithDescr || variable instanceof IImmutableVariable) {
                TType type = variable.getType();
                if (type instanceof ArrayType) {
                    type = ((ArrayType)type).getElementType();
                }
                if (type == null || !type.getBindingKey().startsWith("Ljava/lang/Object;")) {
                    
                    String description;
                    if (variable instanceof ImmutableTypeVariableWithDescr) {
                        description = ((ImmutableTypeVariableWithDescr)variable).getDescription();
                    } else {
                        description = "used as unchangable type argument";
                    }
                    InferTypeConstraintsModel.setAccessSet(variable, new AccessSet(description, variable));
                }
                
            }
        }
    }

    private static boolean hasSuperfluousMethods(TypeHierarchy typeHierarchy, AccessSet accessSet) throws JavaModelException {
        TCType existingType = typeHierarchy.getFittingExistingType(accessSet);
        return existingType == null || existingType != typeHierarchy.getOriginalSelectedType();
    }
    

}
