/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc.hierarchy;

import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.IMethodBinding;

/**
 * A wrapper for {@link IMethodBinding} instances. Used for all methods in the hierarchy
 * of the selected type.
 * <p>
 * The wrapper is necessary to provide support for updating bindings when type
 * occurences are changed in method parameters or return types in the hierarchy.
 * </p>
 * <p>
 * The index of the method declaration within the declaring type is used to identify the method
 * instead of the binding key if possible (that means if it is a source binding)
 * </p>
 * @author Hannes Kegel
 *
 */
public class BindingWrapper {
    private IMethodBinding methodBinding;
    private IMethod originalMethod;
    private int index = -1;

    /**
     * The constructor that automatically determines the index.
     * 
     * @param methodBinding the binding
     * @throws JavaModelException
     */
    public BindingWrapper(IMethodBinding methodBinding) throws JavaModelException {
        this.methodBinding = methodBinding;

        IMethod methodElement = (IMethod)methodBinding.getJavaElement();
        if (methodElement != null) { // null for synthetic methods
            IMethod[] methods = ((IType)methodBinding.getDeclaringClass().getJavaElement()).getMethods();
            for (int i = 0; i < methods.length; i++) {
                if (methods[i].equals(methodElement)) {
                    index = i;
                    return;
                }
            }
            // this is a workaround for a bug with raw overides
            for (int i = 0; i < methods.length; i++) {
                if (methods[i].isSimilar(methodElement)) {
                    index = i;
                    return;
                }
            }
            
            // cannot be a source type, thus no changes to binding key
        }
    }

    /**
     * The constructor that gets a pre-defined index. Used by the binding updater.
     * 
     * @param methodBinding the binding
     * @param index the method index within the declaring type
     */
    public BindingWrapper(IMethodBinding methodBinding, int index) {
        this.methodBinding = methodBinding;
        this.index = index;
    }

    /**
     * Retrieves the binding.
     * @return the binding
     */
    public IMethodBinding getMethodBinding() {
        return methodBinding;
    }
    
    /**
     * Update the method binding if it was changed.
     * 
     * @param methodBinding the new binding
     */
    public void setUpdatedMethodBinding(IMethodBinding methodBinding) {
        //LogUtil.debugOut(this, "setting method binding: " + methodBinding.getKey() + "; old: " + this.methodBinding.getKey());
        if (originalMethod == null) {
            originalMethod = (IMethod)this.methodBinding.getJavaElement();
        }
        this.methodBinding = methodBinding;
    }
    
    /**
     * Retrieves the IMethod of the original (non-updated) binding
     * 
     * @return the original IMethod
     */
    public IMethod getOriginalMethod() {
        return originalMethod == null ? (IMethod)methodBinding.getJavaElement() : originalMethod;
    }

    /**
     * Gets the binding key of the declaring type.
     * 
     * @return the type binding key
     */
    public String getTypeKey() {
        return methodBinding.getDeclaringClass().getKey();
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        return methodBinding.toString();
    }

    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        final int PRIME = 31;
        int result = 1;
        result = PRIME * result + (index == -1 && methodBinding != null ? methodBinding.getKey().hashCode() : index);
        result = PRIME * result + ((methodBinding == null) ? 0 : getTypeKey().hashCode());
        return result;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object obj) {        
        if (this == obj) {
            return true;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final BindingWrapper other = (BindingWrapper) obj;
        if (index != other.index) {
            return false;
        }
        if (!getTypeKey().equals(other.getTypeKey())) {
            return false;
        }
        if (index == -1 && (methodBinding == null || !methodBinding.getKey().equals(other.methodBinding.getKey()))) {
            return false;
        }
        return true;
    }
}
