/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc.hierarchy;

import java.util.HashSet;

import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.internal.typeCalc.BindingSet;

/**
 * A TCType in the hierarchy of the selected type that represent a 
 * new type to be created by IT.
 * 
 * @author Hannes Kegel
 *
 */
public class NewTCType extends TCType {
    private String typeName;
    private BindingSet allMethods; 
    private BindingSet declaredMethods = new BindingSet();
    
    private boolean isInterface = true;
    
    /**
     * The single constructor.
     * 
     * @param allMethods all methods this type should declare or inherit
     * @param typeHierarchy the hierarchy
     * @param baseTCType the existing type this type will be based on 
     */
    public NewTCType(BindingSet allMethods, TypeHierarchy typeHierarchy, ExistingTCType baseTCType) {
        super(baseTCType.baseTypeBinding);
        this.allMethods = allMethods;
        interfaces = new HashSet<TCType>();
        declaredMethods.addAll(allMethods);
    }
    
    /**
     * Sets the qualified name for this new type.
     * 
     * @param typeName the fully qualified name
     */
    public void setTypeName(String typeName) {
        this.typeName = typeName;
    }

    /**
     * Retrieves all declared methods of this new type.
     * 
     * @return BindingSet
     */
    public BindingSet getDeclaredMethods() {
        return declaredMethods;
    }
    
    /**
     * Retrieves the type parameters of the erasure of the type this
     * type is based on.
     * 
     * @param importRewrite the import rewrite to get simple type names
     * @return the parameters including brackets.
     */
    public String getErasureTypeParameters(ImportRewrite importRewrite) {
        return ParameterizedTypeUtil.getErasureTypeParameters(baseTypeBinding, importRewrite);
    }
    
    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#getTypeName()
     */
    @Override
    public String getTypeName() {
        return typeName;
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#getAllMethods()
     */
    @Override
    public BindingSet getAllMethods() {
        return allMethods;
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#setSuperType(org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType, org.eclipse.ltk.core.refactoring.RefactoringStatus)
     */
    @Override
    public void setSuperType(TCType superType, RefactoringStatus status) throws JavaModelException {
        super.setSuperType(superType, status);
        isInterface = false;
        BindingSet allSuperMethods = superType.getAllMethods();
        declaredMethods.removeAllSuperMethods(allSuperMethods);
    }
    

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#isInterface()
     */
    @Override
    public boolean isInterface() {
        return isInterface;
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#addInterface(org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType, org.eclipse.ltk.core.refactoring.RefactoringStatus)
     */
    @Override
    public boolean addInterface(TCType newInterface, RefactoringStatus status) throws JavaModelException {
        if (super.addInterface(newInterface, status)) {
            declaredMethods.removeAllSuperMethods(newInterface.getAllMethods());
            return true;
        }
        return false;
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#canBeMovedInHierarchy()
     */
    @Override
    public boolean canBeMovedInHierarchy() {
        return true;
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#isModifiable()
     */
    @Override
    public boolean isModifiable() {
        return true;
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#hasOtherDeclaredMembers()
     */
    @Override
    public boolean hasOtherDeclaredMembers() {
        return false;
    }
    
}
