/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoj.riwd;

import java.lang.reflect.InvocationTargetException;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.ltk.core.refactoring.Change;
import org.eclipse.ltk.core.refactoring.IValidationCheckResultQuery;
import org.eclipse.ltk.core.refactoring.PerformChangeOperation;
import org.eclipse.ltk.core.refactoring.RefactoringCore;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.eclipse.ltk.ui.refactoring.RefactoringUI;
import org.eclipse.swt.widgets.Display;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElementVisitor;
import org.intoJ.inferType3.internal.util.ConversionUtil;
import org.intoJ.inferType3.internal.util.InternalCalculationException;
import org.intoj.riwd.RiwdRefactoring.Status;

/**
 * This runnable can be used to start Infer Type from a different plugin without worrying
 * about the eclipse refactoring framework.
 * 
 * @author Hannes Kegel
 *
 */
public class ExternalRiwdRunnable implements IRunnableWithProgress {
    
    private RiwdRefactoring refactoring;
    private final boolean checkOnly;
    
    private RefactoringStatus refactoringStatus;

    
    /**
     * The general constructor with bindings and type offsets.
     * 
     * @param binding the binding to be refactored
     * @param declaringClass its declaring type
     * @param typeStartPosition the start position of the type (not of the identifier)
     * @param typeVariableKeys if the selection is a type argument of the binding, this list
     * holds all binding keys to the selected (nested) type. Otherwise null.
     * @param clearCache whether to clear the cache (should be set when used the first time)
     * @param changeTypeArgs if type arguments should be changed if necessary
     * @param noAbstractClasses if set, no abstract classes will be generated for case 3
     * @param checkOnly if set, the refactoring will only check if the selection can be generalized. Call
     * {@link #run(IProgressMonitor)} and afterwards {@link #wasApplicable()}.
     * @throws InternalCalculationException 
     */
    public ExternalRiwdRunnable(IType type, boolean clearCache, boolean checkOnly) throws Exception {
        this.checkOnly = checkOnly;
		ConversionUtil.Result result = ConversionUtil.findBinding(type);
        SelectedElement selectedElement = new SelectedElement(result.getBinding(), result.getTypeStartPos(), null);
        
        refactoring = new RiwdRefactoring(selectedElement, clearCache, false);
    }
        
    
    /* (non-Javadoc)
     * @see org.eclipse.jface.operation.IRunnableWithProgress#run(org.eclipse.core.runtime.IProgressMonitor)
     */
    public void run(final IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
        try {
            monitor.beginTask("", 200);
            refactoringStatus = refactoring.checkInitialConditions(new SubProgressMonitor(monitor, 170));
            if (!checkOnly && !refactoringStatus.hasFatalError()) {
                refactoringStatus.merge(refactoring.checkFinalConditions(new SubProgressMonitor(monitor, 10)));
                if (!refactoringStatus.hasFatalError()) {
                    final Change change = refactoring.createChange(new SubProgressMonitor(monitor, 5));
                    if (change != null) {
                        change.initializeValidationData(new SubProgressMonitor(monitor, 5));                        
                        final PerformChangeOperation operation = RefactoringUI.createUIAwareChangeOperation(change);
                        operation.setUndoManager(RefactoringCore.getUndoManager(), "Replace Inheritance With Delegation");
                        final CoreException[] exHolder = new CoreException[1];
                        Display.getDefault().syncExec(
                                new Runnable() {
                                  public void run(){
                                      try {
                                        operation.run(new SubProgressMonitor(monitor, 10));
                                    } catch (CoreException e) {
                                        exHolder[0] = e;
                                    }
                                  }
                                });
                        if (exHolder[0] != null) {
                            throw exHolder[0];
                        }
                    }
                }

            }
        } catch (OperationCanceledException e) {
            throw e;
        } catch (Throwable e) {
            e.printStackTrace();
            throw new InvocationTargetException(e);
        } finally {
            monitor.done();
        }
    }

    /**
     * Retrieves the error message if there was an error. Otherwise returns null.
     * 
     * @return the message
     */
    public String getErrorMessage() {
        return refactoringStatus.getMessageMatchingSeverity(RefactoringStatus.FATAL);
    }

    /**
     * Checks if there would be or were changes.
     * 
     * @return true if the refactorings can generalize the selection.
     */
    public boolean wasApplicable() {
        return !refactoringStatus.hasFatalError();
    }

    /**
     * Retrieves the merged status of the operation.
     * 
     * @return a RefactoringStatus
     */
    public RefactoringStatus getRefactoringStatus() {
        return refactoringStatus;
    }
    
    public Status getCommonStatus() {
    	return refactoring.getCommonStatus();
    }
    
    public Status getForwardingStatus() {
    	return refactoring.getForwardingStatus();
    }

    public Status getDelegationStatus() {
    	return refactoring.getDelegationStatus();
    }

	public Manipulation.Stats getStats() {
		return refactoring.getStats();
	}
    
}
