/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoj.riwd;

import java.io.IOException;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.WorkingCopyOwner;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.core.dom.SuperConstructorInvocation;
import org.eclipse.jdt.internal.corext.dom.Bindings;
import org.eclipse.jdt.internal.corext.refactoring.structure.ASTNodeSearchUtil;
import org.eclipse.jdt.internal.corext.refactoring.structure.CompilationUnitRewrite;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TypeEnvironment;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2;
import org.eclipse.ltk.core.refactoring.Change;
import org.eclipse.ltk.core.refactoring.Refactoring;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.creation.Collector;
import org.intoJ.inferType3.internal.constraintModel.util.InferTypeGmlHandler;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.constraintModel.variables.BindingKeyProvider;
import org.intoJ.inferType3.internal.constraintModel.variables.DeclaringTypeVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutabilityReason;
import org.intoJ.inferType3.internal.constraintModel.variables.ImmutableTypeVariableWithDescr;
import org.intoJ.inferType3.internal.constraintModel.variables.SuperVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ThisVariable;
import org.intoJ.inferType3.internal.typeCalc.AccessSet;
import org.intoJ.inferType3.internal.typeCalc.AccessSetCalculation;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ExistingTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;
import org.intoJ.inferType3.internal.util.Cache;
import org.intoJ.inferType3.internal.util.GmlWriter;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

public final class RiwdRefactoring extends Refactoring {

    public static final String MSG_INTERNAL_ERROR = "Internal error";
    public static final String MSG_CREATING_CHANGES = "Creating changes";
    public static final String MSG_CALCULATING_ACCESS_SETS = "Calculating access sets";

    public static final int ERROR_CODE_TYPE_ARGUMENT_MODIFICATION_NECESSARY = 1;

    private final WorkingCopyOwner owner = new WorkingCopyOwner() { };

    private SelectedElement selectedElement;
    private TypeHierarchy typeHierarchy;

    private InferTypeConstraintsModel model;
    private Manipulation manipulation;
        
    private RefactoringStatus finalStatus;
    
    private boolean clearCache;

    private Iterable<BindingWrapper> checkedMembers;
    private Iterable<ExistingTCType> checkedInterfaces;
    
    private Status forwardingStatus = new Status();
    private Status delegationStatus = new Status();
    private Status commonStatus = new Status();
    
    private boolean performForwarding = true;
    
    private Set<TType> requiredInterfaces = new HashSet<TType>();
	private final boolean reportInitialStatus;		

    
    
    /**
     * Creates a new Infer Type processor. 
     * 
     * The cache will be cleared.
     * 
     * @param declarationElement the selected declaration element
     */
    public RiwdRefactoring(final SelectedElement declarationElement) {
        this(declarationElement, true, true);
    }

    /**
     * Creates a new Infer Type processor. 
     * 
     * @param declarationElement the selected declaration element
     * @param clearCache if the cache should be cleared
     */
    public RiwdRefactoring(final SelectedElement declarationElement, boolean clearCache, boolean reportInitialStatus) {
        this.selectedElement = declarationElement;
        this.clearCache = clearCache;
		this.reportInitialStatus = reportInitialStatus;
        if (clearCache) {
            Cache.clearAll();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void finalize() throws Throwable {
        resetWorkingCopies();
    }

    /**
     * Resets all working copies.
     */
    private void resetWorkingCopies() {
        final ICompilationUnit[] units = JavaCore.getWorkingCopies(owner);
        for (final ICompilationUnit unit : units) {
            try {
                unit.discardWorkingCopy();
            } catch (Exception exception) {
                // Do nothing
            }
        }
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.ltk.core.refactoring.Refactoring#checkFinalConditions(org.eclipse.core.runtime.IProgressMonitor)
     */
    @Override
    public final RefactoringStatus checkFinalConditions(final IProgressMonitor monitor) throws CoreException, OperationCanceledException {
        Assert.isNotNull(monitor);
        if (finalStatus != null) {
            return finalStatus;
        }
        finalStatus = new RefactoringStatus();

        try {
            monitor.beginTask("", 100);
            monitor.setTaskName(RiwdRefactoring.MSG_CREATING_CHANGES);

            if (!finalStatus.hasFatalError()) {
                // create all changes necessary to introduce the new types
                // do this here to be able to report errors with the refactoring status
            	manipulation.calculateChanges(new SubProgressMonitor(monitor, 85), finalStatus, 
            			getValidationContext(), performForwarding, 
            			performForwarding ? forwardingStatus.getRequiredSupertype() : delegationStatus.getRequiredSupertype(), 
            					requiredInterfaces, model);
            	
            	checkStatus(finalStatus);            	
            }
        } catch (InternalCalculationException e) {
            finalStatus.addFatalError(e.getMessage());
        } finally {
            monitor.done();
        }

        return finalStatus;
    }
       
    /* (non-Javadoc)
     * @see org.eclipse.ltk.core.refactoring.Refactoring#checkInitialConditions(org.eclipse.core.runtime.IProgressMonitor)
     */
    @Override
    public final RefactoringStatus checkInitialConditions(final IProgressMonitor monitor) throws CoreException, OperationCanceledException {
        Assert.isNotNull(monitor);
        final RefactoringStatus status = new RefactoringStatus();
        try {
            TypeEnvironment environment = new TypeEnvironment();
            model = new InferTypeConstraintsModel(environment, selectedElement);
            
            ITypeBinding typeBinding = selectedElement.getDeclaredType();
			// build hierarchy of refactored type
            final IType declaredType = (IType) typeBinding.getJavaElement();
            typeHierarchy = new TypeHierarchy(declaredType, true, true);
            TType type = model.createTType(typeBinding);

            // collect affected files and build constraints
                            
            Collector collector = new Collector(owner, model, selectedElement.getBinding().getJavaElement().getJavaProject(), typeHierarchy, true, true, true, true, true, true);
            collector.run(monitor, status);
            
            if (!status.hasFatalError()) {	            
            	if (checkPrerequesites(typeBinding)) {
                	checkStatus(status);            		
            	}
            }
            
            if (!status.hasFatalError()) {
            	AccessSetCalculation.calculateAllSets(model, typeHierarchy, true);            	            	
	            checkRequiredSupertype(type);
            	checkOpenRecursion(typeHierarchy.getOriginalSelectedType(), typeHierarchy.getSubTypes());
            	//TODO: check open recursion for delegation
	            checkSuperConstructors(status, typeBinding, declaredType);	    
	            checkSynchronizedMethods();
            	
            }
            
            if (reportInitialStatus) {
            	checkStatus(status);
            }
        	
        	if (!forwardingStatus.isPossible()) {
        		performForwarding = false;
        	}
            
            
        } catch (InternalCalculationException e) {
            status.addFatalError(e.getMessage());
        } finally {
            if (clearCache) {
                Cache.clearAll();
            }
        }
        return status;
    }

	private void checkSynchronizedMethods() {
		ITypeBinding type = typeHierarchy.getOriginalSelectedType().getBaseTypeBinding();
		boolean synchronizedMethodsInType = hasSynchronizedMethods(type);
		if (!synchronizedMethodsInType) {
			for (ExistingTCType subType : typeHierarchy.getSubTypes()) {
				synchronizedMethodsInType = hasSynchronizedMethods(subType.getBaseTypeBinding());
				if (synchronizedMethodsInType) break;
			}
		}
		if (synchronizedMethodsInType) {
			ITypeBinding superclass = type.getSuperclass();
			while (superclass != null) {
				if (hasSynchronizedMethods(superclass)) {
					commonStatus.addViolation("Synchronized methods split between sub- and superclasses", ImmutabilityReason.SYNCHRONIZED_SPLIT);
					return;
				}
				superclass = superclass.getSuperclass();
			}
		}
	}

	private boolean hasSynchronizedMethods(ITypeBinding type) {
		for (IMethodBinding method : type.getDeclaredMethods()) {
			if (Modifier.isSynchronized(method.getModifiers())) {
				return true;
			}
		}
		return false;
	}

	private void checkStatus(final RefactoringStatus status) {
		if (!commonStatus.isPossible() || !(forwardingStatus.isPossible() || delegationStatus.isPossible())) {
			StringBuilder builder = new StringBuilder();
			builder.append("Refactoring not possible:");
			String descr = commonStatus.getViolationDescription();
			if (descr.length() > 0) {
				builder.append("\n").append(descr).append("\n");
			}
			descr = forwardingStatus.getViolationDescription();
			if (descr.length() > 0) {
				builder.append("Forwarding specific:\n").append(descr).append("\n");
			}
			descr = delegationStatus.getViolationDescription();
			if (descr.length() > 0) {
				builder.append("Delegation specific:\n").append(descr).append("\n");
			}
			status.addFatalError(builder.toString());
		}
	}
    
    private boolean checkPrerequesites(ITypeBinding typeToCheck) {
    	boolean abort = false;
    	
    	if (typeToCheck.isAnnotation() || typeToCheck.isEnum() || typeToCheck.isInterface()) {
            commonStatus.addViolation("The selected type is not a class.", ImmutabilityReason.NO_CLASS);
            abort = true;
    	}
    	    	
    	ITypeBinding superClass = typeToCheck.getSuperclass();
        if (superClass == null || superClass.getQualifiedName().equals("java.lang.Object")) {
            commonStatus.addViolation("The super type of the selected class is Object.", ImmutabilityReason.OBJECT_SUPERTYPE);
            abort = true;
        }
        if (!typeToCheck.isTopLevel()) {
            commonStatus.addViolation("Replace Inheritance With Delegation supports only top-level classes.", ImmutabilityReason.NO_TOP_LEVEL_TYPE);
            abort = true;
        }
        if (superClass != null) {
	        if (Modifier.isAbstract(superClass.getModifiers())) {
	            commonStatus.addViolation("Replace Inheritance With Delegation is not defined for abstract super classes.", ImmutabilityReason.ABSTRACT_SUPERCLASS);
	        }
	        if (isThrowable(superClass)) {
	            commonStatus.addViolation("Replace Inheritance With Delegation does not support refactoring of Throwables.", ImmutabilityReason.THROWABLE);
	        }
        }
        return abort;
    }

    private boolean isThrowable(ITypeBinding type) {
    	if (type == null) {
    		return false;
    	}
    	if (type.getQualifiedName().equals("java.lang.Throwable")) {
    		return true;
    	}
		return isThrowable(type.getSuperclass());
	}
    

	private void checkSuperConstructors(final RefactoringStatus status,
			ITypeBinding typeBinding, final IType declaredType)
			throws JavaModelException {
		final CompilationUnitRewrite sourceRewrite = new CompilationUnitRewrite(declaredType.getCompilationUnit());
		final AbstractTypeDeclaration declaration = ASTNodeSearchUtil.getAbstractTypeDeclarationNode(declaredType, sourceRewrite.getRoot());
		if (declaration != null) {            	
			SuperCallVisitor superCallVisitor = new SuperCallVisitor();
			declaration.accept(superCallVisitor);
			
			boolean samePackage = declaredType.getPackageFragment().equals(((IType)typeBinding.getSuperclass().getJavaElement()).getPackageFragment());             
			for (SuperConstructorInvocation superConstructorInvocation : superCallVisitor.getSuperConstructorInvocationList()) {
				IMethodBinding methodBinding = superConstructorInvocation.resolveConstructorBinding();
				if (!isAccessible(methodBinding, samePackage)) {
					forwardingStatus.addViolation("Super constructor " + methodBinding.getKey() + " not accessible.", ImmutabilityReason.CONSTRUCTOR_NOT_ACCESSIBLE);             		            								
				}
			}
			if (superCallVisitor.getConstructorWithoutSuperList().size() > 0 || 
					superCallVisitor.getSuperConstructorInvocationList().size() == 0) 
			{
				for (IMethodBinding methodBinding : typeBinding.getSuperclass().getDeclaredMethods()) {
					if (methodBinding.isConstructor() && methodBinding.getParameterTypes().length == 0 && !isAccessible(methodBinding, samePackage)) {
						forwardingStatus.addViolation("Super default constructor not accessible.", ImmutabilityReason.CONSTRUCTOR_NOT_ACCESSIBLE);             		            								
		            	break;
					}
				}
			}
			
			manipulation = new Manipulation(typeHierarchy, sourceRewrite, superCallVisitor, commonStatus);
		} else {
			status.addFatalError("Could not find declaration of selected class.");             		            	
		}
	}

    private boolean isAccessible(IMethodBinding methodBinding,
			boolean samePackage) {
		return samePackage || Modifier.isPublic(methodBinding.getModifiers());
	}

	private void checkOpenRecursion(ExistingTCType type, Set<ExistingTCType> subTypes) {
    	Map<String, Boolean> methodToOverridden = new HashMap<String, Boolean>();
    	
    	TType tType = model.createTType(type.getBaseTypeBinding());
    	

    	for (ITypeConstraint2 constraint : model.getTypeConstraints()) {
			if (constraint.getLeft() instanceof ThisVariable) {
				TType leftType = constraint.getLeft().getType();
				if (leftType != null && (isSubclass(tType.getSuperclass().getTypeDeclaration(), leftType.getTypeDeclaration()) ||
						tType.getSuperclass().getErasure().getBindingKey().equals(leftType.getErasure().getBindingKey()))) {
					// this call in one of the super classes

					AccessSet accessSet = InferTypeConstraintsModel.getAccessSet(constraint.getLeft());
					if (accessSet != null) {
						for (BindingWrapper bindingWrapper : accessSet.getMembers()) {
							if (isOverridden(type, subTypes, methodToOverridden, bindingWrapper.getMethodBinding())) {
								forwardingStatus.addViolation("Open recursion", ImmutabilityReason.OPEN_RECURSION);
							}							
						}
					}
				}
			}
		}
	}

	private Boolean isOverridden(ExistingTCType type,
			Set<ExistingTCType> subTypes,
			Map<String, Boolean> methodToOverriden, IMethodBinding methodBinding) {
		Boolean overriden = methodToOverriden.get(methodBinding.getKey());
		if (overriden == null) {
			overriden = Boolean.FALSE;
			
		    IMethodBinding overridingMethod= Bindings.findMethodInType(type.getBaseTypeBinding(), 
		    		methodBinding.getName(), methodBinding.getParameterTypes());
		    if (overridingMethod != null) {
		    	overriden = Boolean.TRUE;
		    }
		    if (!overriden) {
		    	for (ExistingTCType existingTCType : subTypes) {
		            overridingMethod= Bindings.findMethodInType(existingTCType.getBaseTypeBinding(), 
		            		methodBinding.getName(), methodBinding.getParameterTypes());
		            if (overridingMethod != null) {
		            	overriden = Boolean.TRUE;
		            	break;
		            }
				}
		    }
		    methodToOverriden.put(methodBinding.getKey(), overriden);
		}
		return overriden;
	}

	private void checkRequiredSupertype(TType type) {
		Set<TType> searchedTypes = new HashSet<TType>();
    	for (ExistingTCType existingTCType : typeHierarchy.getSubTypes()) {
			searchedTypes.add(model.createTType(existingTCType.getBaseTypeBinding()));
		}
    	searchedTypes.add(type);
    	
    	Map<TType, Set<ImmutableTypeVariableWithDescr>> requiredSupertypes = new HashMap<TType, Set<ImmutableTypeVariableWithDescr>>();
    	Map<TType, Set<ImmutableTypeVariableWithDescr>> requiredSupertypesForwarding = new HashMap<TType, Set<ImmutableTypeVariableWithDescr>>();
    	for (ITypeConstraint2 constraint : model.getTypeConstraints()) {
			TType leftType = constraint.getLeft().getType();
			if (leftType != null && searchedTypes.contains(leftType.getErasure())) {
				if (!(constraint.getRight() instanceof DeclaringTypeVariable)) {
					TType rightType = constraint.getRight().getType();
					if (rightType != null) {
						rightType = rightType.getErasure();
					}
					if (constraint.getRight() instanceof ImmutableTypeVariableWithDescr) {
						ImmutableTypeVariableWithDescr immutableCv = (ImmutableTypeVariableWithDescr) constraint.getRight();
						boolean forwardingOnly = false;
						if ((immutableCv.getReason().contains(ImmutabilityReason.PACKAGE_METHOD) || 
								immutableCv.getReason().contains(ImmutabilityReason.PROTECTED_METHOD)) && 
								(constraint.getLeft() instanceof ThisVariable || constraint.getLeft() instanceof SuperVariable)) 
						{
							if ((isSubclass(leftType, type.getSuperclass())
									&& getPackage(leftType).equals(getPackage(type.getSuperclass()))))
							{
								immutableCv = null;
							} else {
								forwardingOnly = true;
							}
							
						}
						if (immutableCv != null && immutableCv.getReason().equals(EnumSet.of(ImmutabilityReason.STATIC_METHOD))) {
							immutableCv = null;
						}
						if (immutableCv != null) {
							if (forwardingOnly) {
								addVariable(requiredSupertypesForwarding, (ImmutableTypeVariableWithDescr)constraint.getRight(), rightType);								
							} else {
								addVariable(requiredSupertypes, (ImmutableTypeVariableWithDescr)constraint.getRight(), rightType);
								
							}
						}
					} else {
						addVariable(requiredSupertypes, new ImmutableTypeVariableWithDescr(constraint.getRight().getType(), "Assignment to supertype.", null, ImmutabilityReason.ASSIGNMENT_TO_SUPERCLASS), 
								rightType);								
					}

				}
			}
		}
    	
		Set<TType> superTypes = new HashSet<TType>();
		
		TType superclass = type.getSuperclass();
		while (superclass != null && !superclass.isJavaLangObject()) {
			superTypes.add(superclass.getTypeDeclaration());
			superclass = superclass.getSuperclass();
		}

		for (ConstraintVariable2 cv : model.getConstraintVariables()) {
			if (cv instanceof ImmutableTypeVariableWithDescr) {
				ImmutableTypeVariableWithDescr immCv = (ImmutableTypeVariableWithDescr) cv;
				if (immCv.getReason().contains(ImmutabilityReason.INSTANCE_OF) || immCv.getReason().contains(ImmutabilityReason.CAST)) {
					if (superTypes.contains(immCv.getType())) {
						addVariable(requiredSupertypes, immCv, immCv.getType());														
					}
				}
			}			
		}
    	
    	checkRequiredTypes(type, commonStatus, requiredSupertypes);
    	checkRequiredTypes(type, forwardingStatus, requiredSupertypesForwarding);
    	
    	if (forwardingStatus.getRequiredSupertype() == null || (commonStatus.getRequiredSupertype() != null && 
    			isSubclass(commonStatus.getRequiredSupertype(), forwardingStatus.getRequiredSupertype()))) {
    		forwardingStatus.setRequiredSupertype(commonStatus.getRequiredSupertype());
    	}
    	delegationStatus.setRequiredSupertype(commonStatus.getRequiredSupertype());

	}

	private void checkRequiredTypes(TType type, Status status,
			Map<TType, Set<ImmutableTypeVariableWithDescr>> requiredSupertypes) {
		Map.Entry<TType, Set<ImmutableTypeVariableWithDescr>> requiredSupertypeToDescr = null;
    	for (Map.Entry<TType, Set<ImmutableTypeVariableWithDescr>> superType : requiredSupertypes.entrySet()) {
			if (isSubclass(type, superType.getKey()) && (requiredSupertypeToDescr == null ||  
					isSubclass(superType.getKey(), requiredSupertypeToDescr.getKey()))) 
			{
				requiredSupertypeToDescr = superType;
			} else if (superType.getKey().isInterface()) {
				requiredInterfaces.add(superType.getKey());
			}
		}
    	if (requiredSupertypeToDescr != null && requiredSupertypeToDescr.getKey().isJavaLangObject()) {
    		requiredSupertypeToDescr = null;
    	}
    	
		if (requiredSupertypeToDescr != null) {
			if (requiredSupertypeToDescr.getKey().getErasure().getBindingKey().equals(type.getSuperclass().getErasure().getBindingKey())) {
		    	for (ImmutableTypeVariableWithDescr variable : requiredSupertypeToDescr.getValue()) {
			    	status.addViolation(variable.getDescription(), variable.getReason());
				}
		    }
			status.setRequiredSupertype(requiredSupertypeToDescr.getKey().getTypeDeclaration());
		}
	}

	private void addVariable(
			Map<TType, Set<ImmutableTypeVariableWithDescr>> requiredSupertypes,
			ImmutableTypeVariableWithDescr cv, TType type) {
		Set<ImmutableTypeVariableWithDescr> descrSet = requiredSupertypes.get(type);
		if (descrSet == null) {
			descrSet = new HashSet<ImmutableTypeVariableWithDescr>();
			requiredSupertypes.put(type, descrSet);
		}
		descrSet.add(cv);
	}

	private String getPackage(TType type) {
		String str = type.getPrettySignature();
		int firstBracketPos =  str.indexOf('<');
		if (firstBracketPos > -1) {
			str = str.substring(0, firstBracketPos);
		}
		int lastDotPos =  str.lastIndexOf('.');
		if (lastDotPos == -1) {
			return "";
		} else {
			return str.substring(0, lastDotPos);
		}
	}

	private boolean isSubclass(TType type, TType possibleSupertype) {
		TType superClass = type.getSuperclass();
		if (superClass != null) {
			if (superClass.getErasure().getBindingKey().equals(possibleSupertype.getErasure().getBindingKey())) {
				return true;
			}
			return isSubclass(superClass, possibleSupertype);
		}
		return false;
	}

	/* (non-Javadoc)
     * @see org.eclipse.ltk.core.refactoring.Refactoring#createChange(org.eclipse.core.runtime.IProgressMonitor)
     */
    @Override
    public final Change createChange(final IProgressMonitor monitor) throws CoreException, OperationCanceledException {
        Assert.isNotNull(monitor);
        try {
            monitor.beginTask("", 1);
            monitor.setTaskName("Modifing files");
            return manipulation.getAllChanges();
        } finally {
            monitor.done();
        }
    }

    /**
     * Retrieves the type hierarchy of the selected declaration element. Call after
     * checkInitialConditions has run. 
     * 
     * @return the type hierarchy
     */
    public TypeHierarchy getTypeHierarchy() {
        return typeHierarchy;
    }

    public Status getForwardingStatus() {
		return forwardingStatus;
	}

	public Status getDelegationStatus() {
		return delegationStatus;
	}

	public Status getCommonStatus() {
		return commonStatus;
	}
	
	public Manipulation.Stats getStats() {
		return manipulation == null ? null : manipulation.getStats();
	}
	
	public boolean isPerformForwarding() {
		return performForwarding;
	}

	public void setPerformForwarding(boolean performForwarding) {
		this.performForwarding = performForwarding;
	}

	/**
     * Used to set the additional public members as well as complete interfaces 
     * that will be included in the initial new type.
     * 
     * @param checkedMembers the additional methods.
     * @param checkedInterfaces the interfaces that the new type should be a sub type of (at least).
     */
    public void setCheckedMembers(Iterable<BindingWrapper> checkedMembers, Iterable<ExistingTCType> checkedInterfaces) {
        this.checkedMembers = checkedMembers;
        this.checkedInterfaces = checkedInterfaces;
    }
    /**
     * Create a gml file from the current constraint graph. Call after
     * checkInitialConditions has run. More information about the file format can be found 
     * in {@link GmlWriter}
     * 
     * @param fileName the desired file name 
     * @param elementEdges if the graph should contain edges that show the relationship between element 
     * variables and its parametrized parent types 
     * @param onlyTraversed if only the part of the graph that will be traversed for IT should be exported
     * @throws IOException 
     * @throws JavaModelException 
     */
    public void exportGraph(String fileName, boolean elementEdges, boolean onlyTraversed) throws JavaModelException, IOException {
        InferTypeGmlHandler.writeGml(fileName, elementEdges, onlyTraversed, model, typeHierarchy);
    }

    /* (non-Javadoc)
     * @see org.eclipse.ltk.core.refactoring.Refactoring#getName()
     */
    public String getName() {
        return "Replace Inheritance With Delegation";
    }

    public static class Status {
    	private boolean possible = true;
    	private Set<String> violationDescriptions = new HashSet<String>();
    	private EnumSet<ImmutabilityReason> reasons = EnumSet.noneOf(ImmutabilityReason.class);
    	private TType requiredSupertype;
    	
		public void addViolation(String description,
				ImmutabilityReason reason) 
		{
			addViolation(description, EnumSet.of(reason));
		}

		public void addViolation(String description,
				EnumSet<ImmutabilityReason> newReasons) {
			reasons.addAll(newReasons);
			violationDescriptions.add(description);	
			possible = false;	
		}

		public boolean isPossible() {
			return possible;
		}

		public EnumSet<ImmutabilityReason> getReasons() {
			return reasons;
		}

		public TType getRequiredSupertype() {
			return requiredSupertype;
		}

		public void setRequiredSupertype(TType requiredSupertype) {
			this.requiredSupertype = requiredSupertype;
		}

		public String getViolationDescription() {
			StringBuilder ret = new StringBuilder();
			for (Iterator<String> it = violationDescriptions.iterator(); it.hasNext();) {
				ret.append(it.next());
				if (it.hasNext()) {
					ret.append("\n");
				}				
			}
			return ret.toString();
		}
    }
}
